import itertools
import json
import logging
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Any, Collection, Dict, Generic, Iterator, List, Optional, Set, Tuple, Type, TypeVar, Union, cast

import requests

import sec_certs.constants as constants
import sec_certs.helpers as helpers
import sec_certs.parallel_processing as cert_processing
from sec_certs.config.configuration import config
from sec_certs.dataset.cpe import CPEDataset
from sec_certs.dataset.cve import CVEDataset
from sec_certs.model.cpe_matching import CPEClassifier
from sec_certs.sample.certificate import Certificate
from sec_certs.sample.cpe import CPE
from sec_certs.serialization.json import ComplexSerializableType, serialize

logger = logging.getLogger(__name__)

T = TypeVar("T")
CertSubType = TypeVar("CertSubType", bound=Certificate)
DatasetSubType = TypeVar("DatasetSubType", bound="Dataset")


class Dataset(Generic[CertSubType], ABC):
    def __init__(
        self,
        certs: Dict[str, CertSubType],
        root_dir: Path,
        name: str = "dataset name",
        description: str = "dataset_description",
    ):
        self._root_dir = root_dir
        self.timestamp = datetime.now()
        self.sha256_digest = "not implemented"
        self.name = name
        self.description = description
        self.certs = certs

    @property
    def root_dir(self) -> Path:
        return self._root_dir

    @root_dir.setter
    def root_dir(self, new_dir: Union[str, Path]) -> None:
        new_dir = Path(new_dir)
        new_dir.mkdir(exist_ok=True)
        self._root_dir = new_dir

    @property
    def web_dir(self) -> Path:
        return self.root_dir / "web"

    @property
    def auxillary_datasets_dir(self) -> Path:
        return self.root_dir / "auxillary_datasets"

    @property
    def cpe_dataset_path(self) -> Path:
        return self.auxillary_datasets_dir / "cpe_dataset.json"

    @property
    def cve_dataset_path(self) -> Path:
        return self.auxillary_datasets_dir / "cve_dataset.json"

    @property
    def nist_cve_cpe_matching_dset_path(self) -> Path:
        return self.auxillary_datasets_dir / "nvdcpematch-1.0.json"

    @property
    def json_path(self) -> Path:
        return self.root_dir / (self.name + ".json")

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, Certificate):
            raise TypeError(
                f"You attempted to check if {type(item)} is member of {type(self)}, but only {type(Certificate)} are allowed to be members."
            )
        return item.dgst in self.certs

    def __iter__(self) -> Iterator[CertSubType]:
        yield from self.certs.values()

    def __getitem__(self, item: str) -> CertSubType:
        return self.certs.__getitem__(item.lower())

    def __setitem__(self, key: str, value: CertSubType):
        self.certs.__setitem__(key.lower(), value)

    def __len__(self) -> int:
        return len(self.certs)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Dataset):
            return NotImplemented
        return self.certs == other.certs

    def __str__(self) -> str:
        return str(type(self).__name__) + ":" + self.name + ", " + str(len(self)) + " certificates"

    def to_dict(self) -> Dict[str, Any]:
        return {
            "timestamp": self.timestamp,
            "sha256_digest": self.sha256_digest,
            "name": self.name,
            "description": self.description,
            "n_certs": len(self),
            "certs": list(self.certs.values()),
        }

    @classmethod
    def from_dict(cls: Type[DatasetSubType], dct: Dict) -> DatasetSubType:
        certs = {x.dgst: x for x in dct["certs"]}
        dset = cls(certs, Path("../"), dct["name"], dct["description"])
        if len(dset) != (claimed := dct["n_certs"]):
            logger.error(
                f"The actual number of certs in dataset ({len(dset)}) does not match the claimed number ({claimed})."
            )
        return dset

    @classmethod
    def from_json(cls: Type[DatasetSubType], input_path: Union[str, Path]) -> DatasetSubType:
        dset = cast("DatasetSubType", ComplexSerializableType.from_json(input_path))
        dset.root_dir = Path(input_path).parent.absolute()
        dset.set_local_paths()
        return dset

    def set_local_paths(self) -> None:
        raise NotImplementedError("Not meant to be implemented by the base class.")

    @abstractmethod
    def get_certs_from_web(self) -> None:
        raise NotImplementedError("Not meant to be implemented by the base class.")

    @abstractmethod
    def convert_all_pdfs(self) -> None:
        raise NotImplementedError("Not meant to be implemented by the base class.")

    @abstractmethod
    def download_all_pdfs(self, cert_ids: Optional[Set[str]] = None) -> None:
        raise NotImplementedError("Not meant to be implemented by the base class.")

    @staticmethod
    def _download_parallel(urls: Collection[str], paths: Collection[Path], prune_corrupted: bool = True) -> None:
        exit_codes = cert_processing.process_parallel(
            helpers.download_file, list(zip(urls, paths)), config.n_threads, unpack=True
        )
        n_successful = len([e for e in exit_codes if e == requests.codes.ok])
        logger.info(f"Successfully downloaded {n_successful} files, {len(exit_codes) - n_successful} failed.")

        for url, e in zip(urls, exit_codes):
            if e != requests.codes.ok:
                logger.error(f"Failed to download {url}, exit code: {e}")

        if prune_corrupted is True:
            for p in paths:
                if p.exists() and p.stat().st_size < constants.MIN_CORRECT_CERT_SIZE:
                    logger.error(f"Corrupted file at: {p}")
                    p.unlink()

    def _prepare_cpe_dataset(self, download_fresh_cpes: bool = False, init_lookup_dicts: bool = True) -> CPEDataset:
        logger.info("Preparing CPE dataset.")
        if not self.auxillary_datasets_dir.exists():
            self.auxillary_datasets_dir.mkdir(parents=True)

        if not self.cpe_dataset_path.exists() or download_fresh_cpes is True:
            cpe_dataset = CPEDataset.from_web(self.cpe_dataset_path, init_lookup_dicts)
            cpe_dataset.to_json(str(self.cpe_dataset_path))
        else:
            cpe_dataset = CPEDataset.from_json(str(self.cpe_dataset_path))

        return cpe_dataset

    def _prepare_cve_dataset(
        self, download_fresh_cves: bool = False, use_nist_cpe_matching_dict: bool = True
    ) -> CVEDataset:
        logger.info("Preparing CVE dataset.")
        if not self.auxillary_datasets_dir.exists():
            self.auxillary_datasets_dir.mkdir(parents=True)

        if not self.cve_dataset_path.exists() or download_fresh_cves is True:
            cve_dataset = CVEDataset.from_web()
            cve_dataset.to_json(str(self.cve_dataset_path))
        else:
            cve_dataset = CVEDataset.from_json(str(self.cve_dataset_path))

        cve_dataset.build_lookup_dict(use_nist_cpe_matching_dict, self.nist_cve_cpe_matching_dset_path)
        return cve_dataset

    def _compute_cpe_matches(
        self, download_fresh_cpes: bool = False
    ) -> Tuple[CPEClassifier, CPEDataset, Optional[CVEDataset]]:
        def filter_condition(cpe: CPE) -> bool:
            """
            Filters out very weak CPE matches that don't improve our database.
            """
            if (
                cpe.title
                and (cpe.version == "-" or cpe.version == "*")
                and not any(char.isdigit() for char in cpe.title)
            ):
                return False
            elif (
                not cpe.title
                and cpe.item_name
                and (cpe.version == "-" or cpe.version == "*")
                and not any(char.isdigit() for char in cpe.item_name)
            ):
                return False
            return True

        logger.info("Computing heuristics: Finding CPE matches for certificates")
        cpe_dset = self._prepare_cpe_dataset(download_fresh_cpes, init_lookup_dicts=False)
        cve_dset = None
        if not cpe_dset.was_enhanced_with_vuln_cpes:
            cve_dset = self._prepare_cve_dataset(download_fresh_cves=False)
            cpe_dset.enhance_with_cpes_from_cve_dataset(cve_dset)  # this also calls build_lookup_dicts() on cpe_dset
        else:
            cpe_dset.build_lookup_dicts()

        clf = CPEClassifier(config.cpe_matching_threshold, config.cpe_n_max_matches)
        clf.fit([x for x in cpe_dset if filter_condition(x)])

        cert: CertSubType
        for cert in helpers.tqdm(self, desc="Predicting CPE matches with the classifier"):
            cert.compute_heuristics_cpe_match(clf)

        return clf, cpe_dset, cve_dset

    @serialize
    def compute_cpe_heuristics(self) -> Tuple[CPEClassifier, CPEDataset, Optional[CVEDataset]]:
        return self._compute_cpe_matches()

    def to_label_studio_json(self, output_path: Union[str, Path]) -> None:
        lst = []
        for cert in [x for x in cast(Iterator[Certificate], self) if x.heuristics.cpe_matches]:
            dct = {"text": cert.label_studio_title}
            candidates = [x[1].title for x in cert.heuristics.cpe_matches]
            candidates += ["No good match"] * (config.cc_cpe_max_matches - len(candidates))
            options = ["option_" + str(x) for x in range(1, 21)]
            dct.update({o: c for o, c in zip(options, candidates)})
            lst.append(dct)

        with Path(output_path).open("w") as handle:
            json.dump(lst, handle, indent=4)

    @serialize
    def load_label_studio_labels(self, input_path: Union[str, Path]) -> None:
        with Path(input_path).open("r") as handle:
            data = json.load(handle)

        cpe_dset = self._prepare_cpe_dataset()

        logger.info("Translating label studio matches into their CPE representations and assigning to certificates.")
        for annotation in helpers.tqdm(
            [x for x in data if "verified_cpe_match" in x], desc="Translating label studio matches"
        ):
            match_keys = annotation["verified_cpe_match"]
            match_keys = [match_keys] if isinstance(match_keys, str) else match_keys["choices"]
            match_keys = [x.lstrip("$") for x in match_keys]
            predicted_annotations = [annotation[x] for x in match_keys if annotation[x] != "No good match"]

            cpes: Set[CPE] = set()
            for x in predicted_annotations:
                if x not in cpe_dset.title_to_cpes:
                    print(f"Error: {x} not in dataset")
                else:
                    to_update = cpe_dset.title_to_cpes[x]
                    if to_update and not cpes:
                        cpes = to_update
                    elif to_update and cpes:
                        cpes.update(to_update)

            # distinguish between FIPS and CC
            if "\n" in annotation["text"]:
                cert_name = annotation["text"].split("\nModule name: ")[1].split("\n")[0]
            else:
                cert_name = annotation["text"]

            certs = self.get_certs_from_name(cert_name)

            for c in certs:
                c.heuristics.verified_cpe_matches = {x.uri for x in cpes if x is not None} if cpes else None

    def get_certs_from_name(self, name: str) -> List[CertSubType]:
        raise NotImplementedError("Not meant to be implemented by the base class.")

    def enrich_automated_cpes_with_manual_labels(self) -> None:
        """
        Prior to CVE matching, it is wise to expand the database of automatic CPE matches with those that were manually assigned.
        """
        for cert in cast(Iterator[Certificate], self):
            if not cert.heuristics.cpe_matches and cert.heuristics.verified_cpe_matches:
                cert.heuristics.cpe_matches = cert.heuristics.verified_cpe_matches
            elif cert.heuristics.cpe_matches and cert.heuristics.verified_cpe_matches:
                cert.heuristics.cpe_matches = set(cert.heuristics.cpe_matches).union(
                    set(cert.heuristics.verified_cpe_matches)
                )

    @serialize
    def compute_related_cves(
        self,
        download_fresh_cves: bool = False,
        use_nist_cpe_matching_dict: bool = True,
        cve_dset: Optional[CVEDataset] = None,
    ) -> None:
        logger.info("Retrieving related CVEs to verified CPE matches")
        if download_fresh_cves or not cve_dset:
            cve_dset = self._prepare_cve_dataset(download_fresh_cves, use_nist_cpe_matching_dict)

        self.enrich_automated_cpes_with_manual_labels()
        cpe_rich_certs = [x for x in cast(Iterator[Certificate], self) if x.heuristics.cpe_matches]

        if not cpe_rich_certs:
            logger.error(
                "No certificates with verified CPE match detected. You must run dset.manually_verify_cpe_matches() first. Returning."
            )
            return

        relevant_cpes = set(itertools.chain.from_iterable([x.heuristics.cpe_matches for x in cpe_rich_certs]))
        cve_dset.filter_related_cpes(relevant_cpes)

        cert: Certificate
        for cert in helpers.tqdm(cpe_rich_certs, desc="Computing related CVES"):
            cert.compute_heuristics_related_cves(cve_dset)

        n_vulnerable = len([x for x in cpe_rich_certs if x.heuristics.related_cves])
        n_vulnerabilities = sum([len(x.heuristics.related_cves) for x in cpe_rich_certs if x.heuristics.related_cves])
        logger.info(
            f"In total, we identified {n_vulnerabilities} vulnerabilities in {n_vulnerable} vulnerable certificates."
        )



import logging
import os
import tempfile
from itertools import groupby
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

from bs4 import BeautifulSoup, NavigableString
from graphviz import Digraph

from sec_certs import constants as constants
from sec_certs import helpers as helpers
from sec_certs import parallel_processing as cert_processing
from sec_certs.config.configuration import config
from sec_certs.dataset.dataset import Dataset
from sec_certs.dataset.fips_algorithm import FIPSAlgorithmDataset
from sec_certs.helpers import fips_dgst
from sec_certs.model.dependency_finder import DependencyFinder
from sec_certs.sample.fips import FIPSCertificate
from sec_certs.serialization.json import ComplexSerializableType, serialize

logger = logging.getLogger(__name__)


class FIPSDataset(Dataset[FIPSCertificate], ComplexSerializableType):
    def __init__(
        self,
        certs: Dict[str, FIPSCertificate],
        root_dir: Path,
        name: str = "dataset name",
        description: str = "dataset_description",
    ):
        super().__init__(certs, root_dir, name, description)
        self.keywords: Dict[str, Dict] = {}
        self.algorithms: Optional[FIPSAlgorithmDataset] = None
        self.new_files = 0

    @property
    def results_dir(self) -> Path:
        return self.root_dir / "results"

    @property
    def policies_dir(self) -> Path:
        return self.root_dir / "security_policies"

    @property
    def fragments_dir(self) -> Path:
        return self.root_dir / "fragments"

    @property
    def algs_dir(self) -> Path:
        return self.web_dir / "algorithms"

    # After web scan, there should be a FIPSCertificate object created for every entry
    @property
    def successful_web_scan(self) -> bool:
        return all(self.certs) and all(cert.web_scan for cert in self.certs.values() if cert is not None)

    @property
    def successful_pdf_scan(self) -> bool:
        return all(cert.pdf_scan for cert in self.certs.values() if cert is not None)

    def get_certs_from_name(self, module_name: str) -> List[FIPSCertificate]:
        return [crt for crt in self if crt.web_scan.module_name == module_name]

    def find_empty_pdfs(self) -> Tuple[List, List]:
        missing = []
        not_available = []
        for i in self.certs:
            if not (self.policies_dir / f"{i}.pdf").exists():
                missing.append(i)
            elif os.path.getsize(self.policies_dir / f"{i}.pdf") < constants.FIPS_NOT_AVAILABLE_CERT_SIZE:
                not_available.append(i)
        return missing, not_available

    @serialize
    def pdf_scan(self, redo: bool = False) -> None:
        logger.info("Entering PDF scan.")

        self.fragments_dir.mkdir(parents=True, exist_ok=True)

        keywords = cert_processing.process_parallel(
            FIPSCertificate.find_keywords,
            [cert for cert in self.certs.values() if not cert.pdf_scan.keywords or redo],
            config.n_threads,
            use_threading=False,
            progress_bar_desc="Scanning PDF files",
        )
        for keyword, cert in keywords:
            self.certs[cert.dgst].pdf_scan.keywords = keyword

    def match_algs(self) -> Dict[str, int]:
        output = {}
        for cert in self.certs.values():
            # if the pdf has not been processed, no matching can be done
            if not cert.pdf_scan.keywords or not cert.state.txt_state:
                continue

            output[cert.dgst] = FIPSCertificate.match_web_algs_to_pdf(cert)
            cert.heuristics.unmatched_algs = output[cert.dgst]

        output = {k: v for k, v in output.items() if v != 0}
        return output

    def download_all_pdfs(self, cert_ids: Optional[Set[str]] = None) -> None:
        sp_paths, sp_urls = [], []
        self.policies_dir.mkdir(exist_ok=True)
        if cert_ids is None:
            raise RuntimeError("You need to provide cert ids to FIPS download PDFs functionality.")
        for cert_id in cert_ids:
            if not (self.policies_dir / f"{cert_id}.pdf").exists() or (
                fips_dgst(cert_id) in self.certs and not self.certs[fips_dgst(cert_id)].state.txt_state
            ):
                sp_urls.append(
                    f"https://csrc.nist.gov/CSRC/media/projects/cryptographic-module-validation-program/documents/security-policies/140sp{cert_id}.pdf"
                )
                sp_paths.append(self.policies_dir / f"{cert_id}.pdf")
        logger.info(f"downloading {len(sp_urls)} module pdf files")
        cert_processing.process_parallel(
            FIPSCertificate.download_security_policy,
            list(zip(sp_urls, sp_paths)),
            config.n_threads,
            progress_bar_desc="Downloading PDF files",
        )
        self.new_files += len(sp_urls)

    def download_all_htmls(self, cert_ids: Set[str]) -> List[str]:
        html_paths, html_urls = [], []
        new_files = []
        self.web_dir.mkdir(exist_ok=True)
        for cert_id in cert_ids:
            if not (self.web_dir / f"{cert_id}.html").exists():
                html_urls.append(
                    f"https://csrc.nist.gov/projects/cryptographic-module-validation-program/certificate/{cert_id}"
                )
                html_paths.append(self.web_dir / f"{cert_id}.html")
                new_files.append(cert_id)

        logger.info(f"downloading {len(html_urls)} module html files")
        failed = cert_processing.process_parallel(
            FIPSCertificate.download_html_page,
            list(zip(html_urls, html_paths)),
            config.n_threads,
            progress_bar_desc="Downloading HTML files",
        )
        failed = [c for c in failed if c]

        self.new_files += len(html_urls)
        if len(failed) != 0:
            logger.info(f"Download failed for {len(failed)} files. Retrying...")
            cert_processing.process_parallel(
                FIPSCertificate.download_html_page,
                failed,
                config.n_threads,
                progress_bar_desc="Downloading HTML files again",
            )
        return new_files

    @serialize
    def convert_all_pdfs(self) -> None:
        logger.info("Converting FIPS sample reports to .txt")
        tuples = [
            (cert, self.policies_dir / f"{cert.cert_id}.pdf", self.policies_dir / f"{cert.cert_id}.pdf.txt")
            for cert in self.certs.values()
            if not cert.state.txt_state and (self.policies_dir / f"{cert.cert_id}.pdf").exists()
        ]
        cert_processing.process_parallel(
            FIPSCertificate.convert_pdf_file, tuples, config.n_threads, progress_bar_desc="Converting to txt"
        )

    def prepare_dataset(self, test: Optional[Path] = None, update: bool = False) -> Set[str]:
        if test:
            html_files = [test]
        else:
            html_files = [
                Path("fips_modules_active.html"),
                Path("fips_modules_historical.html"),
                Path("fips_modules_revoked.html"),
            ]
            helpers.download_file(
                "https://csrc.nist.gov/projects/cryptographic-module-validation-program/validated-modules/search?SearchMode=Advanced&CertificateStatus=Active&ValidationYear=0",
                Path(self.web_dir / "fips_modules_active.html"),
            )
            helpers.download_file(
                "https://csrc.nist.gov/projects/cryptographic-module-validation-program/validated-modules/search?SearchMode=Advanced&CertificateStatus=Historical&ValidationYear=0",
                Path(self.web_dir / "fips_modules_historical.html"),
            )
            helpers.download_file(
                "https://csrc.nist.gov/projects/cryptographic-module-validation-program/validated-modules/search?SearchMode=Advanced&CertificateStatus=Revoked&ValidationYear=0",
                Path(self.web_dir / "fips_modules_revoked.html"),
            )

        # Parse those files and get list of currently processable files (always)
        cert_ids: Set[str] = set()
        for f in html_files:
            cert_ids |= self._get_certificates_from_html(self.web_dir / f, update)

        return cert_ids

    def download_neccessary_files(self, cert_ids: Set[str]) -> None:
        self.download_all_htmls(cert_ids)
        self.download_all_pdfs(cert_ids)

    def _get_certificates_from_html(self, html_file: Path, update: bool = False) -> Set[str]:
        logger.info(f"Getting certificate ids from {html_file}")
        with open(html_file, "r", encoding="utf-8") as handle:
            html = BeautifulSoup(handle.read(), "html5lib")

        table = [x for x in html.find(id="searchResultsTable").tbody.contents if x != "\n"]
        entries: Set[str] = set()

        for entry in table:
            if isinstance(entry, NavigableString):
                continue
            cert_id = entry.find("a").text
            if cert_id not in entries:
                entries.add(cert_id)

        return entries

    @serialize
    def web_scan(self, cert_ids: Set[int], redo: bool = False) -> None:
        logger.info("Entering web scan.")
        for cert_id in cert_ids:
            dgst = fips_dgst(cert_id)
            self.certs[dgst] = FIPSCertificate.html_from_file(
                self.web_dir / f"{cert_id}.html",
                FIPSCertificate.State(
                    (self.policies_dir / str(cert_id)).with_suffix(".pdf"),
                    (self.web_dir / str(cert_id)).with_suffix(".html"),
                    (self.fragments_dir / str(cert_id)).with_suffix(".txt"),
                    False,
                    None,
                    False,
                ),
                self.certs.get(dgst),
                redo=redo,
            )

    @classmethod
    def from_web_latest(cls) -> "FIPSDataset":
        with tempfile.TemporaryDirectory() as tmp_dir:
            dset_path = Path(tmp_dir) / "fips_latest_dataset.json"
            logger.info("Downloading the latest FIPS dataset.")
            helpers.download_file(config.fips_latest_snapshot, dset_path)
            dset: FIPSDataset = cls.from_json(dset_path)
            logger.info(
                "The dataset with %s certs and %s algorithms.",
                len(dset),
                len(dset.algorithms) if dset.algorithms is not None else 0,
            )
            logger.info("The dataset does not contain the results of the dependency analysis - calculating them now...")
            dset.finalize_results()
            return dset

    def set_local_paths(self) -> None:
        cert: FIPSCertificate
        for cert in self.certs.values():
            cert.set_local_paths(self.policies_dir, self.web_dir, self.fragments_dir)

    @serialize
    def get_certs_from_web(
        self,
        test: Optional[Path] = None,
        no_download_algorithms: bool = False,
        update: bool = False,
        redo_web_scan=False,
    ) -> None:
        """Downloads HTML search pages, parses them, populates the dataset,
        and performs `web-scan` - extracting information from CMVP pages for
        each certificate.

        Args:
            test (Optional[Path], optional): Path to dataset used in testing. Defaults to None.
            no_download_algorithms (bool, optional): Whether to reuse CAVP algorithm dataset. Defaults to False.
            update (bool, optional): Whether to update dataset with new entries. Defaults to False.
            redo_web_scan (bool, optional): Whether to redo the `web-scan` functionality. Defaults to False.
        """
        logger.info("Downloading required html files")

        self.web_dir.mkdir(parents=True, exist_ok=True)
        self.policies_dir.mkdir(exist_ok=True)
        self.algs_dir.mkdir(exist_ok=True)

        # Download files containing all available module certs (always)
        cert_ids = self.prepare_dataset(test, update)

        if not no_download_algorithms:
            aset = FIPSAlgorithmDataset({}, Path(self.root_dir / "web" / "algorithms"), "algorithms", "sample algs")
            aset.get_certs_from_web()
            logger.info(f"Finished parsing. Have algorithm dataset with {len(aset)} algorithm numbers.")

            self.algorithms = aset

        logger.info("Downloading certificate html and security policies")
        self.download_neccessary_files(cert_ids)

        self.web_scan(cert_ids, redo=redo_web_scan)

    @serialize
    def deprocess(self) -> None:
        # TODO think of a better way to make resulting json smaller and easier to share
        logger.info(
            "Removing 'heuristics' field. This dataset can be used to be uploaded and later downloaded using latest_snapshot() or something"
        )
        cert: FIPSCertificate
        for cert in self.certs.values():
            cert.heuristics = FIPSCertificate.FIPSHeuristics(dict(), [], 0)

        self.match_algs()

    @serialize
    def extract_certs_from_tables(self, high_precision: bool) -> List[Path]:
        """
        Function that extracts algorithm IDs from tables in security policies files.
        :return: list of files that couldn't have been decoded
        """
        logger.info("Entering table scan.")
        result = cert_processing.process_parallel(
            FIPSCertificate.analyze_tables,
            [
                (cert, high_precision)
                for cert in self.certs.values()
                if cert is not None and (not cert.state.tables_done or high_precision) and cert.state.txt_state
            ],
            config.n_threads // 4,  # tabula already processes by parallel, so
            # it's counterproductive to use all threads
            use_threading=False,
            progress_bar_desc="Searching tables",
        )

        not_decoded = [cert.state.sp_path for done, cert, _ in result if done is False]
        for state, cert, algorithms in result:
            certificate = self.certs[cert.dgst]
            certificate.state.tables_done = state
            certificate.pdf_scan.algorithms += algorithms
        return not_decoded

    def remove_algorithms_from_extracted_data(self) -> None:
        for cert in self.certs.values():
            cert.remove_algorithms()

    def unify_algorithms(self) -> None:
        for certificate in self.certs.values():
            new_algorithms: List[Dict] = []
            united_algorithms = [
                x
                for x in (
                    (certificate.web_scan.algorithms if certificate.web_scan.algorithms is not None else [])
                    + certificate.pdf_scan.algorithms
                )
                if x != {"Certificate": []}
            ]
            for algorithm in united_algorithms:
                if isinstance(algorithm, dict):
                    new_algorithms.append(algorithm)
                else:
                    new_algorithms.append({"Certificate": [algorithm]})
            certificate.heuristics.algorithms = new_algorithms

            # returns True if candidates should _not_ be matched

    def _compare_certs(self, current_certificate: FIPSCertificate, other_id: str) -> bool:
        other_dgst = fips_dgst(other_id)
        other_cert = self.certs[other_dgst]

        if (
            current_certificate.web_scan.date_validation is None
            or other_cert is None
            or other_cert.web_scan.date_validation is None
        ):
            raise RuntimeError("Building of the dataset probably failed - this should not be happening.")

        cert_first = current_certificate.web_scan.date_validation[0]
        cert_last = current_certificate.web_scan.date_validation[-1]
        conn_first = other_cert.web_scan.date_validation[0]
        conn_last = other_cert.web_scan.date_validation[-1]

        return (
            cert_first.year - conn_first.year > config.year_difference_between_validations
            and cert_last.year - conn_last.year > config.year_difference_between_validations
            or cert_first.year < conn_first.year
        )

    def _remove_false_positives_for_cert(self, current_cert: FIPSCertificate) -> None:
        if current_cert.heuristics.keywords is None:
            raise RuntimeError("Dataset was probably not built correctly - this should not be happening.")
        for rule in current_cert.heuristics.keywords["rules_cert_id"]:
            matches = current_cert.heuristics.keywords["rules_cert_id"][rule]
            current_cert.heuristics.keywords["rules_cert_id"][rule] = [
                cert_id
                for cert_id in matches
                if self._validate_id(current_cert, cert_id.replace("Cert.", "").replace("cert.", "").lstrip("#CA0 "))
                and cert_id != current_cert.cert_id
            ]

    @staticmethod
    def _match_with_algorithm(processed_cert: FIPSCertificate, cert_candidate_id: str) -> bool:
        for cert_alg in processed_cert.heuristics.algorithms:
            for certificate in cert_alg["Certificate"]:
                curr_id = "".join(filter(str.isdigit, certificate))
                if curr_id == cert_candidate_id:
                    return False
        return True

    def _validate_id(self, processed_cert: FIPSCertificate, cert_candidate_id: str) -> bool:
        candidate_dgst = fips_dgst(cert_candidate_id)
        if candidate_dgst not in self.certs or not cert_candidate_id.isdecimal():
            return False

        # "< number" still needs to be used, because of some old certs being revalidated
        if int(cert_candidate_id) < config.smallest_certificate_id_to_connect or self._compare_certs(
            processed_cert, cert_candidate_id
        ):
            return False

        if self.algorithms is None:
            raise RuntimeError("Dataset was probably not built correctly - this should not be happening.")

        if cert_candidate_id not in self.algorithms.certs:
            return True

        if not FIPSDataset._match_with_algorithm(processed_cert, cert_candidate_id):
            return False

        algs = self.algorithms.certs[cert_candidate_id]
        for current_alg in algs:
            if current_alg.vendor is None or processed_cert.web_scan.vendor is None:
                raise RuntimeError("Dataset was probably not built correctly - this should not be happening.")

            if FIPSCertificate.get_compare(processed_cert.web_scan.vendor) == FIPSCertificate.get_compare(
                current_alg.vendor
            ):
                return False
        return True

    def validate_results(self) -> None:
        """
        Function that validates results and finds the final connection output
        """

        def pdf_lookup(cert):
            return set(
                filter(
                    lambda x: x,
                    map(
                        lambda cid: "".join(filter(str.isdigit, cid)),
                        cert.heuristics.keywords["rules_cert_id"].values(),
                    ),
                )
            )

        def web_lookup(cert):
            return set(
                filter(lambda x: x, map(lambda cid: "".join(filter(str.isdigit, cid)), cert.web_scan.mentioned_certs))
            )

        current_cert: FIPSCertificate

        for current_cert in self.certs.values():
            if not current_cert.state.txt_state:
                continue
            self._remove_false_positives_for_cert(current_cert)

        finder = DependencyFinder()
        finder.fit(self.certs, lambda cert: cert.cert_id, pdf_lookup)  # type: ignore

        for dgst in self.certs:
            setattr(self.certs[dgst].heuristics, "st_references", finder.get_references(dgst))

        finder = DependencyFinder()
        finder.fit(self.certs, lambda cert: cert.cert_id, web_lookup)  # type: ignore

        for dgst in self.certs:
            setattr(self.certs[dgst].heuristics, "web_references", finder.get_references(dgst))

    @serialize
    def finalize_results(self, use_nist_cpe_matching_dict: bool = True, perform_cpe_heuristics: bool = True):
        logger.info("Entering 'analysis' and building connections between certificates.")
        self.unify_algorithms()
        self.remove_algorithms_from_extracted_data()
        self.validate_results()
        if perform_cpe_heuristics:
            _, _, cve_dset = self.compute_cpe_heuristics()
            self.compute_related_cves(use_nist_cpe_matching_dict=use_nist_cpe_matching_dict, cve_dset=cve_dset)

    def _highlight_vendor_in_dot(self, dot: Digraph, current_dgst: str, highlighted_vendor: str) -> None:
        current_cert = self.certs[current_dgst]

        if current_cert.web_scan.vendor != highlighted_vendor:
            return

        dot.attr("node", color="red")
        if current_cert.web_scan.status == "Revoked":
            dot.attr("node", color="grey32")
        if current_cert.web_scan.status == "Historical":
            dot.attr("node", color="gold3")

    def _add_colored_node(self, dot: Digraph, current_dgst: str, highlighted_vendor: str) -> None:
        current_cert = self.certs[current_dgst]
        dot.attr("node", color="lightgreen")
        if current_cert.web_scan.status == "Revoked":
            dot.attr("node", color="lightgrey")
        if current_cert.web_scan.status == "Historical":
            dot.attr("node", color="gold")
        self._highlight_vendor_in_dot(dot, current_dgst, highlighted_vendor)
        dot.node(
            str(current_cert.cert_id),
            label=str(current_cert.cert_id) + "&#10;" + current_cert.web_scan.vendor
            if current_cert.web_scan.vendor is not None
            else "" + "&#10;" + (current_cert.web_scan.module_name if current_cert.web_scan.module_name else ""),
        )

    def _get_processed_list(self, connection_list: str, dgst: str) -> List[str]:
        attr = {"pdf": "pdf_scan", "web": "web_scan", "heuristics": "heuristics"}[connection_list]
        return getattr(self.certs[dgst], attr).connections

    def create_dot_graph(
        self,
        output_file_name: str,
        connection_list: str = "heuristics",
        highlighted_vendor: str = "Red Hat®, Inc.",
        show: bool = True,
    ) -> None:
        """
        Function that plots .dot graph of dependencies between certificates
        Certificates with at least one dependency are displayed in "{output_file_name}connections.pdf", remaining
        certificates are displayed in {output_file_name}single.pdf
        :param show: display graph right on screen
        :param highlighted_vendor: vendor whose certificates should be highlighted in red color
        :param output_file_name: prefix to "connections", "connections.pdf", "single" and "single.pdf"
        :param connection_list: 'heuristics', 'web', or 'pdf' - plots a graph from this source
                                default - heuristics
        """
        dot = Digraph(comment="Certificate ecosystem")
        single_dot = Digraph(comment="Modules with no dependencies")
        single_dot.attr("graph", label="Single nodes", labelloc="t", fontsize="30")
        single_dot.attr("node", style="filled")
        dot.attr("graph", label="Dependencies", labelloc="t", fontsize="30")
        dot.attr("node", style="filled")

        keys = 0
        edges = 0

        for key in self.certs:
            cert = self.certs[key]

            if not cert.state.file_status:
                continue

            processed = self._get_processed_list(connection_list, key)

            if processed:
                self._add_colored_node(dot, key, highlighted_vendor)
                keys += 1
            else:
                single_dot.attr("node", color="lightblue")
                self._highlight_vendor_in_dot(dot, key, highlighted_vendor)
                single_dot.node(
                    key,
                    label=str(cert.cert_id) + "\r\n" + cert.web_scan.vendor
                    if cert.web_scan.vendor is not None
                    else "" + ("\r\n" + cert.web_scan.module_name if cert.web_scan.module_name else ""),
                )

        for key in self.certs:
            cert = self.certs[key]

            if not cert.state.file_status:
                continue
            processed = self._get_processed_list(connection_list, key)
            for conn in processed:
                self._add_colored_node(dot, fips_dgst(conn), highlighted_vendor)
                dot.edge(key, conn)
                edges += 1

        logger.info(f"rendering for {connection_list}: {keys} keys and {edges} edges")

        dot.render(self.root_dir / (str(output_file_name) + "_connections"), view=show)
        single_dot.render(self.root_dir / (str(output_file_name) + "_single"), view=show)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "timestamp": self.timestamp,
            "sha256_digest": self.sha256_digest,
            "name": self.name,
            "description": self.description,
            "n_certs": len(self),
            "certs": self.certs,
            "algs": self.algorithms,
        }

    @classmethod
    def from_dict(cls, dct: Dict[str, Any]) -> "FIPSDataset":
        certs = dct["certs"]
        dset = cls(certs, Path("./"), dct["name"], dct["description"])
        dset.algorithms = dct["algs"]
        if len(dset) != (claimed := dct["n_certs"]):
            logger.error(
                f"The actual number of certs in dataset ({len(dset)}) does not match the claimed number ({claimed})."
            )
        return dset

    def group_vendors(self) -> Dict[str, List[str]]:
        vendors: Dict[str, List[str]] = {}
        v = {x.web_scan.vendor.lower() for x in self.certs.values() if x is not None and x.web_scan.vendor is not None}
        v_sorted = sorted(v, key=FIPSCertificate.get_compare)
        for prefix, a in groupby(v_sorted, key=FIPSCertificate.get_compare):
            vendors[prefix] = list(a)

        return vendors

    def plot_graphs(self, show: bool = False) -> None:
        self.create_dot_graph("full_graph", show=show)
        self.create_dot_graph("web_only_graph", "web", show=show)
        self.create_dot_graph("pdf_only_graph", "pdf", show=show)



import pytest

from sec_certs.dataset.cve import CVEDataset
from sec_certs.sample.cve import CVE


class TestCVE:
    @pytest.mark.slow
    @pytest.mark.monitor_test
    def test_from_web(self):
        dset = CVEDataset.from_web()
        assert dset is not None
        assert "CVE-2019-15809" in dset.cves
        assert "CVE-2017-15361" in dset.cves

    def test_from_to_dict(self):
        data = {
            "cve_id": "CVE-1999-0001",
            "vulnerable_cpes": [
                {
                    "uri": "cpe:2.3:o:freebsd:freebsd:1.0:*:*:*:*:*:*:*",
                    "title": None,
                    "start_version": None,
                    "end_version": None,
                }
            ],
            "impact": {
                "_type": "Impact",
                "base_score": 5,
                "severity": "MEDIUM",
                "explotability_score": 10,
                "impact_score": 2.9,
            },
            "published_date": "1999-12-30T05:00:00+00:00",
        }
        cve = CVE.from_dict(data)
        ret = cve.to_dict()
        assert ret == data
        other_cve = CVE.from_dict(ret)
        assert cve == other_cve



import json
import logging
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Dict, Optional, Tuple, Union

import sec_certs.helpers as helpers
from sec_certs.sample.protection_profile import ProtectionProfile

logger = logging.getLogger(__name__)


@dataclass
class ProtectionProfileDataset:
    static_dataset_url: ClassVar[str] = "https://ajanovsky.cz/pp_data_complete_processed.json"

    pps: Dict[Tuple[str, Optional[str]], ProtectionProfile]

    def __iter__(self):
        yield from self.pps.values()

    def __getitem__(self, item: Tuple[str, Optional[str]]) -> ProtectionProfile:
        return self.pps.__getitem__(item)

    def __setitem__(self, key: Tuple[str, Optional[str]], value: ProtectionProfile):
        self.pps.__setitem__(key, value)

    def __contains__(self, key):
        return key in self.pps

    def __len__(self) -> int:
        return len(self.pps)

    @classmethod
    def from_json(cls, json_path: Union[str, Path]):
        with Path(json_path).open("r") as handle:
            data = json.load(handle)
        pps = [ProtectionProfile.from_old_api_dict(x) for x in data.values()]

        dct = {}
        for item in pps:
            if (item.pp_name, item.pp_link) in dct:
                logger.warning(f"Duplicate entry in PP dataset: {(item.pp_name, item.pp_link)}")
            dct[(item.pp_name, item.pp_link)] = item

        return cls(dct)

    @classmethod
    def from_web(cls, store_dataset_path: Optional[Path] = None):
        logger.info(f"Downloading static PP dataset from: {cls.static_dataset_url}")
        if not store_dataset_path:
            tmp = tempfile.TemporaryDirectory()
            store_dataset_path = Path(tmp.name) / "pp_dataset.json"

        helpers.download_file(cls.static_dataset_url, store_dataset_path)
        obj = cls.from_json(store_dataset_path)

        if not store_dataset_path:
            tmp.cleanup()

        return obj



import json
import logging
from pathlib import Path
from typing import List, Optional, Set, Union

import numpy as np

import sec_certs.helpers as helpers
from sec_certs.dataset.cpe import CPEDataset
from sec_certs.sample.common_criteria import CommonCriteriaCert
from sec_certs.sample.fips import FIPSCertificate
from sec_certs.serialization.json import CustomJSONEncoder

logger = logging.getLogger(__name__)


def get_validation_dgsts(filepath: Union[str, Path]) -> Set[str]:
    with Path(filepath).open("r") as handle:
        return set(json.load(handle))


def compute_precision(y: np.ndarray, y_pred: np.ndarray, **kwargs) -> float:
    prec = []
    for true, pred in zip(y, y_pred):
        set_pred = set(pred) if pred else set()
        set_true = set(true) if true else set()
        if set_pred and not set_true:
            prec.append(0.0)
        elif not set_pred and not set_true:
            prec.append(1.0)
        else:
            prec.append(len(set_true.intersection(set_pred)) / len(set_true))
    return np.mean(prec)


def evaluate(
    x_valid: List[Union[CommonCriteriaCert, FIPSCertificate]],
    y_valid: List[Optional[Set[str]]],
    outpath: Optional[Union[Path, str]],
    cpe_dset: CPEDataset,
) -> None:
    y_pred = [x.heuristics.cpe_matches for x in x_valid]
    precision = compute_precision(np.array(y_valid), np.array(y_pred))

    correctly_classified = []
    badly_classified = []
    n_new_certs_with_match = 0
    n_newly_identified = 0

    for cert, predicted_cpes, verified_cpes in zip(x_valid, y_pred, y_valid):
        if not verified_cpes:
            verified_cpes = set()
        verified_cpes_dict = {x: cpe_dset[x].title if cpe_dset[x].title else x for x in verified_cpes}

        if not predicted_cpes:
            predicted_cpes = set()
        predicted_cpes_dict = {x: cpe_dset[x].title if cpe_dset[x].title else x for x in predicted_cpes}

        cert_name = cert.name if isinstance(cert, CommonCriteriaCert) else cert.web_scan.module_name
        vendor = cert.manufacturer if isinstance(cert, CommonCriteriaCert) else cert.web_scan.vendor
        record = {
            "certificate_name": cert_name,
            "vendor": vendor,
            "heuristic version": helpers.compute_heuristics_version(cert_name) if cert_name else None,
            "predicted_cpes": predicted_cpes_dict,
            "manually_assigned_cpes": verified_cpes_dict,
        }

        if verified_cpes.issubset(predicted_cpes):
            correctly_classified.append(record)
        else:
            badly_classified.append(record)

        if not verified_cpes and predicted_cpes:
            n_new_certs_with_match += 1
        n_newly_identified += len(predicted_cpes - verified_cpes)

    results = {
        "Precision": precision,
        "n_new_certs_with_match": n_new_certs_with_match,
        "n_newly_identified": n_newly_identified,
        "correctly_classified": correctly_classified,
        "badly_classified": badly_classified,
    }
    logger.info(
        f"While keeping precision: {precision}, the classifier identified {n_newly_identified} new CPE matches (Found match for {n_new_certs_with_match} certificates that were previously unmatched) compared to baseline."
    )

    if outpath:
        with Path(outpath).open("w") as handle:
            json.dump(results, handle, indent=4, cls=CustomJSONEncoder, sort_keys=True)



import json
import os
import shutil
import tempfile
from datetime import date, datetime
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from unittest import TestCase

import sec_certs.constants as constants
import sec_certs.helpers as helpers
from sec_certs.dataset.common_criteria import CCDataset
from sec_certs.sample.common_criteria import CommonCriteriaCert
from sec_certs.sample.protection_profile import ProtectionProfile


class TestCommonCriteriaOOP(TestCase):
    def setUp(self):
        self.test_data_dir = Path(__file__).parent / "data" / "test_cc_oop"
        self.crt_one = CommonCriteriaCert(
            "active",
            "Access Control Devices and Systems",
            "NetIQ Identity Manager 4.7",
            "NetIQ Corporation",
            "SE",
            {"ALC_FLR.2", "EAL3+"},
            date(2020, 6, 15),
            date(2025, 6, 15),
            "https://www.commoncriteriaportal.org/files/epfiles/Certification%20Report%20-%20NetIQ®%20Identity%20Manager%204.7.pdf",
            "https://www.commoncriteriaportal.org/files/epfiles/ST%20-%20NetIQ%20Identity%20Manager%204.7.pdf",
            "https://www.commoncriteriaportal.org/files/epfiles/Certifikat%20CCRA%20-%20NetIQ%20Identity%20Manager%204.7_signed.pdf",
            "https://www.netiq.com/",
            set(),
            set(),
            None,
            None,
            None,
        )

        self.crt_two = CommonCriteriaCert(
            "active",
            "Access Control Devices and Systems",
            "Magic SSO V4.0",
            "Dreamsecurity Co., Ltd.",
            "KR",
            set(),
            date(2019, 11, 15),
            date(2024, 11, 15),
            "https://www.commoncriteriaportal.org/files/epfiles/KECS-CR-19-70%20Magic%20SSO%20V4.0(eng)%20V1.0.pdf",
            "https://www.commoncriteriaportal.org/files/epfiles/Magic_SSO_V4.0-ST-v1.4_EN.pdf",
            None,
            "https://www.dreamsecurity.com/",
            {
                ProtectionProfile(
                    "Korean National Protection Profile for Single Sign On V1.0",
                    "https://www.commoncriteriaportal.org/files/ppfiles/KECS-PP-0822-2017%20Korean%20National%20PP%20for%20Single%20Sign%20On%20V1.0(eng).pdf",
                )
            },
            set(),
            None,
            None,
            None,
        )

        pp = ProtectionProfile("sample_pp", "https://sample.pp")
        update = CommonCriteriaCert.MaintenanceReport(
            date(1900, 1, 1), "Sample maintenance", "https://maintenance.up", "https://maintenance.up"
        )
        self.fictional_cert = CommonCriteriaCert(
            "archived",
            "Sample category",
            "Sample certificate name",
            "Sample manufacturer",
            "Sample scheme",
            {"Sample security level"},
            date(1900, 1, 2),
            date(1900, 1, 3),
            "https://path.to/report/link",
            "https://path.to/st/link",
            "https://path.to/cert/link",
            "https://path.to/manufacturer/web",
            {pp},
            {update},
            None,
            None,
            None,
        )
        self.template_dataset = CCDataset(
            {self.crt_one.dgst: self.crt_one, self.crt_two.dgst: self.crt_two},
            Path("/fictional/path/to/dataset"),
            "toy dataset",
            "toy dataset description",
        )
        self.template_dataset.timestamp = datetime(2020, 11, 16, hour=17, minute=4, second=14, microsecond=770153)
        self.template_dataset.state.meta_sources_parsed = True

        self.template_report_pdf_hashes = {
            "309ac2fd7f2dcf17": "774c41fbba980191ca40ae610b2f61484c5997417b3325b6fd68b345173bde52",
            "8cf86948f02f047d": "533a5995ef8b736cc48cfda30e8aafec77d285511471e0e5a9e8007c8750203a",
        }
        self.template_target_pdf_hashes = {
            "309ac2fd7f2dcf17": "b9a45995d9e40b2515506bbf5945e806ef021861820426c6d0a6a074090b47a9",
            "8cf86948f02f047d": "3c8614338899d956e9e56f1aa88d90e37df86f3310b875d9d14ec0f71e4759be",
        }

        self.template_report_txt_path = self.test_data_dir / "report_309ac2fd7f2dcf17.txt"
        self.template_target_txt_path = self.test_data_dir / "target_309ac2fd7f2dcf17.txt"

    def test_certificate_input_sanity(self):
        self.assertEqual(
            self.crt_one.report_link,
            "https://www.commoncriteriaportal.org/files/epfiles/Certification%20Report%20-%20NetIQ®%20Identity%20Manager%204.7.pdf",
            "Report link contains some improperly escaped characters.",
        )

    def test_download_and_convert_pdfs(self):
        dset = CCDataset.from_json(self.test_data_dir / "toy_dataset.json")

        with TemporaryDirectory() as td:
            dset.root_dir = Path(td)
            dset.download_all_pdfs()
            dset.convert_all_pdfs()

            actual_report_pdf_hashes = {
                key: helpers.get_sha256_filepath(val.state.report_pdf_path) for key, val in dset.certs.items()
            }
            actual_target_pdf_hashes = {
                key: helpers.get_sha256_filepath(val.state.st_pdf_path) for key, val in dset.certs.items()
            }

            self.assertEqual(
                actual_report_pdf_hashes,
                self.template_report_pdf_hashes,
                "Hashes of downloaded pdfs (sample report) do not the template",
            )
            self.assertEqual(
                actual_target_pdf_hashes,
                self.template_target_pdf_hashes,
                "Hashes of downloaded pdfs (security target) do not match the template",
            )

            self.assertTrue(dset["309ac2fd7f2dcf17"].state.report_txt_path.exists())
            self.assertTrue(dset["309ac2fd7f2dcf17"].state.st_txt_path.exists())

            self.assertAlmostEqual(
                dset["309ac2fd7f2dcf17"].state.st_txt_path.stat().st_size,
                self.template_target_txt_path.stat().st_size,
                delta=1000,
            )

            self.assertAlmostEqual(
                dset["309ac2fd7f2dcf17"].state.report_txt_path.stat().st_size,
                self.template_report_txt_path.stat().st_size,
                delta=1000,
            )

    def test_cert_to_json(self):
        with NamedTemporaryFile("w") as tmp:
            self.fictional_cert.to_json(tmp.name)
            with open(tmp.name) as f:
                out = json.load(f)
            with open(self.test_data_dir / "fictional_cert.json") as f:
                template = json.load(f)
            self.assertEqual(out, template, "Serialized cert differs from cert json template.")

    def test_dataset_to_json(self):
        with NamedTemporaryFile("w") as tmp:
            self.template_dataset.to_json(tmp.name)
            with open(tmp.name, "r") as handle:
                out = json.load(handle)
            with open(self.test_data_dir / "toy_dataset.json") as handle:
                template = json.load(handle)
            self.assertEqual(out, template, "Serialized dataset differs from json template.")

    def test_cert_from_json(self):
        self.assertEqual(
            self.fictional_cert,
            CommonCriteriaCert.from_json(self.test_data_dir / "fictional_cert.json"),
            "The sample serialized from json differs from a template.",
        )

    def test_dataset_from_json(self):
        self.assertEqual(
            self.template_dataset,
            CCDataset.from_json(self.test_data_dir / "toy_dataset.json"),
            "The dataset serialized from json differs from a template.",
        )

    def test_build_empty_dataset(self):
        with TemporaryDirectory() as tmp_dir:
            dset = CCDataset({}, Path(tmp_dir), "sample_dataset", "sample dataset description")
            dset.get_certs_from_web(to_download=False, get_archived=False, get_active=False)
        self.assertEqual(len(dset), 0, "The dataset should contain 0 files.")

    def test_build_dataset(self):
        with TemporaryDirectory() as tmp_dir:
            dataset_path = Path(tmp_dir)
            os.mkdir(dataset_path / "web")
            shutil.copyfile(
                self.test_data_dir / "cc_products_active.csv", dataset_path / "web" / "cc_products_active.csv"
            )
            shutil.copyfile(
                self.test_data_dir / "cc_products_active.html", dataset_path / "web" / "cc_products_active.html"
            )

            dset = CCDataset({}, dataset_path, "sample_dataset", "sample dataset description")
            dset.get_certs_from_web(
                keep_metadata=False, to_download=False, get_archived=False, get_active=True, update_json=False
            )

            self.assertEqual(
                len(os.listdir(dataset_path)),
                0,
                "Meta files (csv, html) were not deleted properly albeit this was explicitly required.",
            )

        self.assertEqual(len(dset), 2, "The dataset should contain 2 files.")

        self.assertTrue(self.crt_one in dset, "The dataset does not contain the template sample.")
        self.assertEqual(dset, self.template_dataset, "The loaded dataset does not match the template dataset.")

    def test_download_csv_html_files(self):
        with TemporaryDirectory() as tmp_dir:
            dataset_path = Path(tmp_dir)
            dset = CCDataset({}, dataset_path, "sample_dataset", "sample dataset description")
            dset.download_csv_html_resources(get_active=True, get_archived=False)

            for x in dset.active_html_tuples:
                self.assertTrue(x[1].exists())
                self.assertGreaterEqual(x[1].stat().st_size, constants.MIN_CC_HTML_SIZE)
            for x in dset.active_csv_tuples:
                self.assertTrue(x[1].exists())
                self.assertGreaterEqual(x[1].stat().st_size, constants.MIN_CC_CSV_SIZE)

    def test_download_pp_dataset(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            self.template_dataset.root_dir = tmp_dir
            self.template_dataset.process_protection_profiles()
            self.assertTrue(self.template_dataset.pp_dataset_path.exists())
            self.assertGreaterEqual(
                self.template_dataset.pp_dataset_path.stat().st_size, constants.MIN_CC_PP_DATASET_SIZE
            )



from pathlib import Path
from typing import ClassVar
from unittest import TestCase

from sec_certs.dataset.fips_mip import MIPDataset
from sec_certs.sample.fips_mip import MIPSnapshot


class TestFIPSMIP(TestCase):
    test_data_dir: ClassVar[Path] = Path(__file__).parent / "data" / "test_fips_mip"
    test_data_dump: ClassVar[Path] = test_data_dir / "fips_mip_2021-02-19T06+01:00.html"

    def test_from_dumps(self):
        dset = MIPDataset.from_dumps(self.test_data_dir)
        self.assertIsNotNone(dset)
        self.assertEqual(len(dset), 3)

    def test_from_dump(self):
        snap = MIPSnapshot.from_dump(self.test_data_dump)
        self.assertIsNotNone(snap)

    def test_from_web(self):
        snap = MIPSnapshot.from_web()
        self.assertIsNotNone(snap)



import re
from enum import Enum

RESPONSE_OK = 200
RETURNCODE_OK = "ok"
RETURNCODE_NOK = "nok"
REQUEST_TIMEOUT = 10

MIN_CORRECT_CERT_SIZE = 5000

MIN_CC_HTML_SIZE = 5000000
MIN_CC_CSV_SIZE = 700000
MIN_CC_PP_DATASET_SIZE = 2500000


class CertFramework(Enum):
    CC = "Common Criteria"
    FIPS = "FIPS"


CPE_VERSION_NA = "-"

FIPS_BASE_URL = "https://csrc.nist.gov"
FIPS_MODULE_URL = "https://csrc.nist.gov/projects/cryptographic-module-validation-program/certificate/"
FIPS_NOT_AVAILABLE_CERT_SIZE = 10000
FIPS_ALG_URL = "https://csrc.nist.gov/projects/cryptographic-algorithm-validation-program/validation-search?searchMode=implementation&page="

FIPS_MIP_STATUS_RE = re.compile(r"^(?P<status>[a-zA-Z ]+?) +\((?P<since>\d{1,2}/\d{1,2}/\d{4})\)$")

TAG_MATCH_COUNTER = "count"
TAG_MATCH_MATCHES = "matches"

TAG_CERT_HEADER_PROCESSED = "cert_header_processed"

TAG_CERT_ID = "cert_id"
TAG_CC_SECURITY_LEVEL = "cc_security_level"
TAG_CC_VERSION = "cc_version"
TAG_CERT_LAB = "cert_lab"
TAG_CERT_ITEM = "cert_item"
TAG_CERT_ITEM_VERSION = "cert_item_version"
TAG_DEVELOPER = "developer"
TAG_REFERENCED_PROTECTION_PROFILES = "ref_protection_profiles"
TAG_HEADER_MATCH_RULES = "match_rules"
TAG_PP_TITLE = "pp_title"
TAG_PP_GENERAL_STATUS = "pp_general_status"
TAG_PP_VERSION_NUMBER = "pp_version_number"
TAG_PP_ID = "pp_id"
TAG_PP_ID_REGISTRATOR = "pp_id_registrator"
TAG_PP_DATE = "pp_date"
TAG_PP_AUTHORS = "pp_authors"
TAG_PP_REGISTRATOR = "pp_registrator"
TAG_PP_REGISTRATOR_SIMPLIFIED = "pp_registrator_simplified"
TAG_PP_SPONSOR = "pp_sponsor"
TAG_PP_EDITOR = "pp_editor"
TAG_PP_REVIEWER = "pp_reviewer"
TAG_KEYWORDS = "keywords"


FILE_ERRORS_STRATEGY = "surrogateescape"
STOP_ON_UNEXPECTED_NUMS = False
APPEND_DETAILED_MATCH_MATCHES = False
LINE_SEPARATOR = " "



from pathlib import Path
from typing import List


def generate_html(ids: List[str], path: Path):
    def generate_entry(certificate_id: str) -> str:
        return f"""
            <tr id="cert-row-0">
                <td class="text-center">
                    <a href="/projects/cryptographic-module-validation-program/certificate/3898" id="cert-number-link-0">{certificate_id}</a>
                </td>
            </tr>
        """

    html_head = """
    <!DOCTYPE html>
    <html lang="en-us" xml:lang="en-us">
    <head>
        <meta charset="utf-8" />
        <title>Cryptographic Module Validation Program | CSRC</title>
        <meta http-equiv="content-type" content="text/html; charset=UTF-8" />
        <meta http-equiv="content-style-type" content="text/css" />
        <meta http-equiv="content-script-type" content="text/javascript" />
        <meta name="viewport" content="width=device-width, initial-scale=1.0" />
        <meta name="msapplication-config" content="/CSRC/Media/images/favicons/browserconfig.xml" />
        <meta name="theme-color" content="#000000" />
        <meta name="google-site-verification" content="xbrnrVYDgLD-Bd64xHLCt4XsPXzUhQ-4lGMj4TdUUTA" />
    </head>
    """
    rows = ""
    for cert_id in ids:
        rows += f"\n{generate_entry(cert_id)}\n"
    html_body = f"""
    <body>
        <table class="table table-striped table-condensed publications-table table-bordered" id="searchResultsTable">
            <thead>
                <tr>
                    <th class="text-center">Certificate Number</th>
                    <th class="text-center">Vendor Name</th>
                    <th class="text-center">Module Name</th>
                    <th class="text-center">Module Type</th>
                    <th class="text-center">Validation Date</th>
                </tr>
            </thead>
            <tbody>
            {rows}
            </tbody>
        </table>
    </body>
    """
    with open(path, "w") as f:
        f.write(f"{html_head}\n{html_body}\n")



import shutil
import tempfile
from pathlib import Path
from typing import ClassVar
from unittest import TestCase

import tests.data.test_cc_heuristics
from sec_certs.dataset.common_criteria import CCDataset
from sec_certs.sample.common_criteria import CommonCriteriaCert


class TestCommonCriteriaTextProcessing(TestCase):
    dataset_json_path: ClassVar[Path] = Path(tests.data.test_cc_heuristics.__path__[0]) / "vulnerable_dataset.json"
    data_dir_path: ClassVar[Path] = dataset_json_path.parent
    tmp_dir: ClassVar[tempfile.TemporaryDirectory]
    cc_dset: CCDataset

    @classmethod
    def setUpClass(cls) -> None:
        cls.tmp_dir = tempfile.TemporaryDirectory()
        shutil.copytree(cls.data_dir_path, cls.tmp_dir.name, dirs_exist_ok=True)

        cls.cc_dset = CCDataset.from_json(Path(cls.tmp_dir.name) / "vulnerable_dataset.json")
        cls.cc_dset.download_all_pdfs()
        cls.cc_dset.convert_all_pdfs()

    def test_extract_frontpage(self):
        subject_cert: CommonCriteriaCert = list(self.cc_dset.certs.values())[0]

        subject_cert.state.st_extract_ok = True
        CommonCriteriaCert.extract_st_pdf_frontpage(subject_cert)
        self.assertTrue(subject_cert.state.st_extract_ok, True)

        subject_cert.state.report_extract_ok = True
        CommonCriteriaCert.extract_report_pdf_frontpage(subject_cert)
        self.assertTrue(subject_cert.state.report_extract_ok, True)

    def test_keyword_extraction(self):
        subject_cert: CommonCriteriaCert = list(self.cc_dset.certs.values())[0]

        subject_cert.state.st_extract_ok = True
        CommonCriteriaCert.extract_st_pdf_keywords(subject_cert)
        self.assertTrue(subject_cert.state.st_extract_ok, True)

        subject_cert.state.report_extract_ok = True
        CommonCriteriaCert.extract_report_pdf_keywords(subject_cert)
        self.assertTrue(subject_cert.state.report_extract_ok, True)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmp_dir.cleanup()



import copy
import operator
from dataclasses import dataclass, field
from datetime import date, datetime
from functools import partial
from pathlib import Path
from typing import Any, Callable, ClassVar, Dict, List, Optional, Set, Tuple, Union

import requests
from bs4 import Tag

from sec_certs import constants as constants
from sec_certs import helpers
from sec_certs.model.cpe_matching import CPEClassifier
from sec_certs.model.dependency_finder import References
from sec_certs.sample.certificate import Certificate, Heuristics, logger
from sec_certs.sample.protection_profile import ProtectionProfile
from sec_certs.serialization.json import ComplexSerializableType
from sec_certs.serialization.pandas import PandasSerializableType

HEADERS = {
    "anssi": helpers.search_only_headers_anssi,
    "bsi": helpers.search_only_headers_bsi,
    "nscib": helpers.search_only_headers_nscib,
    "niap": helpers.search_only_headers_niap,
    "canada": helpers.search_only_headers_canada,
}


class CommonCriteriaCert(
    Certificate["CommonCriteriaCert", "CommonCriteriaCert.CCHeuristics"],
    PandasSerializableType,
    ComplexSerializableType,
):
    cc_url = "http://www.commoncriteriaportal.org"
    empty_st_url = "http://www.commoncriteriaportal.org/files/epfiles/"

    @dataclass(eq=True, frozen=True)
    class MaintenanceReport(ComplexSerializableType):
        """
        Object for holding maintenance reports.
        """

        maintenance_date: Optional[date]
        maintenance_title: Optional[str]
        maintenance_report_link: Optional[str]
        maintenance_st_link: Optional[str]

        def __post_init__(self):
            super().__setattr__("maintenance_report_link", helpers.sanitize_link(self.maintenance_report_link))
            super().__setattr__("maintenance_st_link", helpers.sanitize_link(self.maintenance_st_link))
            super().__setattr__("maintenance_title", helpers.sanitize_string(self.maintenance_title))
            super().__setattr__("maintenance_date", helpers.sanitize_date(self.maintenance_date))

        @classmethod
        def from_dict(cls, dct: Dict) -> "CommonCriteriaCert.MaintenanceReport":
            new_dct = dct.copy()
            new_dct["maintenance_date"] = (
                date.fromisoformat(dct["maintenance_date"])
                if isinstance(dct["maintenance_date"], str)
                else dct["maintenance_date"]
            )
            return super().from_dict(new_dct)

        def __lt__(self, other):
            return self.maintenance_date < other.maintenance_date

    @dataclass(init=False)
    class InternalState(ComplexSerializableType):
        st_download_ok: bool
        report_download_ok: bool
        st_convert_ok: bool
        report_convert_ok: bool
        st_extract_ok: bool
        report_extract_ok: bool
        errors: List[str]

        st_pdf_path: Path
        report_pdf_path: Path
        st_txt_path: Path
        report_txt_path: Path

        def __init__(
            self,
            st_download_ok: bool = True,
            report_download_ok: bool = True,
            st_convert_ok: bool = True,
            report_convert_ok: bool = True,
            st_extract_ok: bool = True,
            report_extract_ok: bool = True,
            errors: Optional[List[str]] = None,
        ):
            self.st_download_ok = st_download_ok
            self.report_download_ok = report_download_ok
            self.st_convert_ok = st_convert_ok
            self.report_convert_ok = report_convert_ok
            self.st_extract_ok = st_extract_ok
            self.report_extract_ok = report_extract_ok
            self.errors = errors if errors else []

        @property
        def serialized_attributes(self) -> List[str]:
            return [
                "st_download_ok",
                "report_download_ok",
                "st_convert_ok",
                "report_convert_ok",
                "st_extract_ok",
                "report_extract_ok",
                "errors",
            ]

        def report_is_ok_to_download(self, fresh: bool = True) -> bool:
            return True if fresh else not self.report_download_ok

        def st_is_ok_to_download(self, fresh: bool = True) -> bool:
            return True if fresh else not self.st_download_ok

        def report_is_ok_to_convert(self, fresh: bool = True) -> bool:
            return self.report_download_ok if fresh else self.report_download_ok and not self.report_convert_ok

        def st_is_ok_to_convert(self, fresh: bool = True) -> bool:
            return self.st_download_ok if fresh else self.st_download_ok and not self.st_convert_ok

        def report_is_ok_to_analyze(self, fresh: bool = True) -> bool:
            if fresh is True:
                return self.report_download_ok and self.report_convert_ok and self.report_extract_ok
            else:
                return self.report_download_ok and self.report_convert_ok and not self.report_extract_ok

        def st_is_ok_to_analyze(self, fresh: bool = True) -> bool:
            if fresh is True:
                return self.st_download_ok and self.st_convert_ok and self.st_extract_ok
            else:
                return self.st_download_ok and self.st_convert_ok and not self.st_extract_ok

    @dataclass
    class PdfData(ComplexSerializableType):
        report_metadata: Optional[Dict[str, Any]] = field(default=None)
        st_metadata: Optional[Dict[str, Any]] = field(default=None)
        report_frontpage: Optional[Dict[str, Dict[str, Any]]] = field(default=None)
        st_frontpage: Optional[Dict[str, Dict[str, Any]]] = field(default=None)
        report_keywords: Optional[Dict[str, Any]] = field(default=None)
        st_keywords: Optional[Dict[str, Any]] = field(default=None)

        def __bool__(self) -> bool:
            return any([x is not None for x in vars(self)])

        @property
        def bsi_data(self) -> Optional[Dict[str, Any]]:
            return self.report_frontpage.get("bsi", None) if self.report_frontpage else None

        @property
        def niap_data(self) -> Optional[Dict[str, Any]]:
            return self.report_frontpage.get("niap", None) if self.report_frontpage else None

        @property
        def nscib_data(self) -> Optional[Dict[str, Any]]:
            return self.report_frontpage.get("nscib", None) if self.report_frontpage else None

        @property
        def canada_data(self) -> Optional[Dict[str, Any]]:
            return self.report_frontpage.get("canada", None) if self.report_frontpage else None

        @property
        def anssi_data(self) -> Optional[Dict[str, Any]]:
            return self.report_frontpage.get("anssi", None) if self.report_frontpage else None

        @property
        def cert_lab(self) -> Optional[List[str]]:
            labs = [
                data["cert_lab"].split(" ")[0].upper()
                for data in [self.bsi_data, self.anssi_data, self.niap_data, self.nscib_data, self.canada_data]
                if data
            ]
            return labs if labs else None

        @property
        def bsi_cert_id(self) -> Optional[str]:
            return self.bsi_data.get("cert_id", None) if self.bsi_data else None

        @property
        def niap_cert_id(self) -> Optional[str]:
            return self.niap_data.get("cert_id", None) if self.niap_data else None

        @property
        def nscib_cert_id(self) -> Optional[str]:
            return self.nscib_data.get("cert_id", None) if self.nscib_data else None

        @property
        def canada_cert_id(self) -> Optional[str]:
            return self.canada_data.get("cert_id", None) if self.canada_data else None

        @property
        def anssi_cert_id(self) -> Optional[str]:
            return self.anssi_data.get("cert_id", None) if self.anssi_data else None

        @property
        def processed_cert_id(self) -> Optional[str]:
            cert_ids = set(
                filter(
                    lambda x: x,
                    {self.bsi_cert_id, self.niap_cert_id, self.nscib_cert_id, self.canada_cert_id, self.anssi_cert_id},
                )
            )
            # Expect only one cert_id in the set above.
            if len(cert_ids) >= 2:
                raise ValueError("More than one cert_id set.")
            elif len(cert_ids) == 1:
                return cert_ids.pop()
            else:
                return None

        @property
        def keywords_rules_cert_id(self) -> Optional[Dict[str, Optional[Dict[str, Dict[str, int]]]]]:
            return self.report_keywords.get("rules_cert_id", None) if self.report_keywords else None

        @property
        def keywords_cert_id(self) -> Optional[str]:
            """
            :return: the most occurring among cert ids captured in keywords scan
            """
            if not self.keywords_rules_cert_id:
                return None

            candidates = [(x, y) for x, y in self.keywords_rules_cert_id.items()]
            candidates = sorted(candidates, key=operator.itemgetter(1), reverse=True)
            return candidates[0][0]

        @property
        def cert_id(self) -> Optional[str]:
            return processed if (processed := self.processed_cert_id) else self.keywords_cert_id

    @dataclass
    class CCHeuristics(Heuristics, ComplexSerializableType):
        extracted_versions: Optional[Set[str]] = field(default=None)
        cpe_matches: Optional[Set[str]] = field(default=None)
        verified_cpe_matches: Optional[Set[str]] = field(default=None)
        related_cves: Optional[Set[str]] = field(default=None)
        cert_lab: Optional[List[str]] = field(default=None)
        cert_id: Optional[str] = field(default=None)
        st_references: References = field(default_factory=References)
        report_references: References = field(default_factory=References)

        @property
        def serialized_attributes(self) -> List[str]:
            return copy.deepcopy(super().serialized_attributes)

    pandas_columns: ClassVar[List[str]] = [
        "dgst",
        "cert_id",
        "name",
        "status",
        "category",
        "manufacturer",
        "scheme",
        "security_level",
        "not_valid_before",
        "not_valid_after",
        "report_link",
        "st_link",
        "manufacturer_web",
        "extracted_versions",
        "cpe_matches",
        "verified_cpe_matches",
        "related_cves",
        "directly_referenced_by",
        "indirectly_referenced_by",
        "directly_referencing",
        "indirectly_referencing",
    ]

    def __init__(
        self,
        status: str,
        category: str,
        name: str,
        manufacturer: Optional[str],
        scheme: str,
        security_level: Union[str, set],
        not_valid_before: Optional[date],
        not_valid_after: Optional[date],
        report_link: str,
        st_link: str,
        cert_link: Optional[str],
        manufacturer_web: Optional[str],
        protection_profiles: Optional[Set[ProtectionProfile]],
        maintenance_updates: Optional[Set[MaintenanceReport]],
        state: Optional[InternalState],
        pdf_data: Optional[PdfData],
        heuristics: Optional[CCHeuristics],
    ):
        super().__init__()

        self.status = status
        self.category = category
        self.name = helpers.sanitize_string(name)

        self.manufacturer = None
        if manufacturer:
            self.manufacturer = helpers.sanitize_string(manufacturer)

        self.scheme = scheme
        self.security_level = helpers.sanitize_security_levels(security_level)
        self.not_valid_before = helpers.sanitize_date(not_valid_before)
        self.not_valid_after = helpers.sanitize_date(not_valid_after)
        self.report_link = helpers.sanitize_link(report_link)
        self.st_link = helpers.sanitize_link(st_link)
        self.cert_link = helpers.sanitize_link(cert_link)
        self.manufacturer_web = helpers.sanitize_link(manufacturer_web)
        self.protection_profiles = protection_profiles
        self.maintenance_updates = maintenance_updates
        self.state = self.InternalState() if not state else state
        self.pdf_data = self.PdfData() if not pdf_data else pdf_data
        self.heuristics: "CommonCriteriaCert.CCHeuristics" = self.CCHeuristics() if not heuristics else heuristics

    @property
    def dgst(self) -> str:
        """
        Computes the primary key of the sample using first 16 bytes of SHA-256 digest
        """
        if not (self.name is not None and self.report_link is not None and self.category is not None):
            raise RuntimeError("Certificate digest can't be computed, because information is missing.")
        return helpers.get_first_16_bytes_sha256(self.category + self.name + self.report_link)

    @property
    def label_studio_title(self) -> str:
        return self.name

    @property
    def pandas_tuple(self) -> Tuple:
        return (
            self.dgst,
            self.heuristics.cert_id,
            self.name,
            self.status,
            self.category,
            self.manufacturer,
            self.scheme,
            self.security_level,
            self.not_valid_before,
            self.not_valid_after,
            self.report_link,
            self.st_link,
            self.manufacturer_web,
            self.heuristics.extracted_versions,
            self.heuristics.cpe_matches,
            self.heuristics.verified_cpe_matches,
            self.heuristics.related_cves,
            self.heuristics.report_references.directly_referenced_by,
            self.heuristics.report_references.indirectly_referenced_by,
            self.heuristics.report_references.directly_referencing,
            self.heuristics.report_references.indirectly_referencing,
        )

    def __str__(self) -> str:
        printed_manufacturer = self.manufacturer if self.manufacturer else "Unknown manufacturer"
        return str(printed_manufacturer) + " " + str(self.name) + " dgst: " + self.dgst

    def merge(self, other: "CommonCriteriaCert", other_source: Optional[str] = None) -> None:
        """
        Merges with other CC sample. Assuming they come from different sources, e.g., csv and html.
        Assuming that html source has better protection profiles, they overwrite CSV info
        On other values (apart from maintenances, see TODO below) the sanity checks are made.
        """
        if self != other:
            logger.warning(
                f"Attempting to merge divergent certificates: self[dgst]={self.dgst}, other[dgst]={other.dgst}"
            )

        for att, val in vars(self).items():
            if not val:
                setattr(self, att, getattr(other, att))
            elif other_source == "html" and att == "protection_profiles":
                setattr(self, att, getattr(other, att))
            elif other_source == "html" and att == "maintenance_updates":
                setattr(self, att, getattr(other, att))
            elif att == "state":
                setattr(self, att, getattr(other, att))
            else:
                if getattr(self, att) != getattr(other, att):
                    logger.warning(
                        f"When merging certificates with dgst {self.dgst}, the following mismatch occured: Attribute={att}, self[{att}]={getattr(self, att)}, other[{att}]={getattr(other, att)}"
                    )

    @classmethod
    def from_dict(cls, dct: Dict) -> "CommonCriteriaCert":
        new_dct = dct.copy()
        new_dct["maintenance_updates"] = set(dct["maintenance_updates"])
        new_dct["protection_profiles"] = set(dct["protection_profiles"])
        new_dct["not_valid_before"] = (
            date.fromisoformat(dct["not_valid_before"])
            if isinstance(dct["not_valid_before"], str)
            else dct["not_valid_before"]
        )
        new_dct["not_valid_after"] = (
            date.fromisoformat(dct["not_valid_after"])
            if isinstance(dct["not_valid_after"], str)
            else dct["not_valid_after"]
        )
        return super(cls, CommonCriteriaCert).from_dict(new_dct)

    @staticmethod
    def _html_row_get_name(cell: Tag) -> str:
        return list(cell.stripped_strings)[0]

    @staticmethod
    def _html_row_get_manufacturer(cell: Tag) -> Optional[str]:
        if lst := list(cell.stripped_strings):
            return lst[0]
        else:
            return None

    @staticmethod
    def _html_row_get_scheme(cell: Tag) -> str:
        return list(cell.stripped_strings)[0]

    @staticmethod
    def _html_row_get_security_level(cell: Tag) -> set:
        return set(cell.stripped_strings)

    @staticmethod
    def _html_row_get_manufacturer_web(cell: Tag) -> Optional[str]:
        for link in cell.find_all("a"):
            if link is not None and link.get("title") == "Vendor's web site" and link.get("href") != "http://":
                return link.get("href")
        return None

    @staticmethod
    def _html_row_get_protection_profiles(cell: Tag) -> set:
        protection_profiles = set()
        for link in list(cell.find_all("a")):
            if link.get("href") is not None and "/ppfiles/" in link.get("href"):
                protection_profiles.add(
                    ProtectionProfile(str(link.contents[0]), CommonCriteriaCert.cc_url + link.get("href"))
                )
        return protection_profiles

    @staticmethod
    def _html_row_get_date(cell: Tag) -> Optional[date]:
        text = cell.get_text()
        extracted_date = datetime.strptime(text, "%Y-%m-%d").date() if text else None
        return extracted_date

    @staticmethod
    def _html_row_get_report_st_links(cell: Tag) -> Tuple[str, str]:
        links = cell.find_all("a")
        assert links[1].get("title").startswith("Certification Report")
        assert links[2].get("title").startswith("Security Target")

        report_link = CommonCriteriaCert.cc_url + links[1].get("href")
        security_target_link = CommonCriteriaCert.cc_url + links[2].get("href")

        return report_link, security_target_link

    @staticmethod
    def _html_row_get_cert_link(cell: Tag) -> Optional[str]:
        links = cell.find_all("a")
        return CommonCriteriaCert.cc_url + links[0].get("href") if links else None

    @staticmethod
    def _html_row_get_maintenance_div(cell: Tag) -> Optional[Tag]:
        divs = cell.find_all("div")
        for d in divs:
            if d.find("div") and d.stripped_strings and list(d.stripped_strings)[0] == "Maintenance Report(s)":
                return d
        return None

    @staticmethod
    def _html_row_get_maintenance_updates(main_div: Tag) -> Set["CommonCriteriaCert.MaintenanceReport"]:
        possible_updates = list(main_div.find_all("li"))
        maintenance_updates = set()
        for u in possible_updates:
            text = list(u.stripped_strings)[0]
            main_date = datetime.strptime(text.split(" ")[0], "%Y-%m-%d").date() if text else None
            main_title = text.split("– ")[1]
            main_report_link = None
            main_st_link = None
            links = u.find_all("a")
            for link in links:
                if link.get("title").startswith("Maintenance Report:"):
                    main_report_link = CommonCriteriaCert.cc_url + link.get("href")
                elif link.get("title").startswith("Maintenance ST"):
                    main_st_link = CommonCriteriaCert.cc_url + link.get("href")
                else:
                    logger.error("Unknown link in Maintenance part!")
            maintenance_updates.add(
                CommonCriteriaCert.MaintenanceReport(main_date, main_title, main_report_link, main_st_link)
            )
        return maintenance_updates

    @classmethod
    def from_html_row(cls, row: Tag, status: str, category: str) -> "CommonCriteriaCert":
        """
        Creates a CC sample from html row
        """

        cells = list(row.find_all("td"))
        if len(cells) != 7:
            logger.error("Unexpected number of cells in CC html row.")
            raise

        name = CommonCriteriaCert._html_row_get_name(cells[0])
        manufacturer = CommonCriteriaCert._html_row_get_manufacturer(cells[1])
        manufacturer_web = CommonCriteriaCert._html_row_get_manufacturer_web(cells[1])
        scheme = CommonCriteriaCert._html_row_get_scheme(cells[6])
        security_level = CommonCriteriaCert._html_row_get_security_level(cells[5])
        protection_profiles = CommonCriteriaCert._html_row_get_protection_profiles(cells[0])
        not_valid_before = CommonCriteriaCert._html_row_get_date(cells[3])
        not_valid_after = CommonCriteriaCert._html_row_get_date(cells[4])
        report_link, st_link = CommonCriteriaCert._html_row_get_report_st_links(cells[0])
        cert_link = CommonCriteriaCert._html_row_get_cert_link(cells[2])
        maintenance_div = CommonCriteriaCert._html_row_get_maintenance_div(cells[0])
        maintenances = (
            CommonCriteriaCert._html_row_get_maintenance_updates(maintenance_div) if maintenance_div else set()
        )

        return cls(
            status,
            category,
            name,
            manufacturer,
            scheme,
            security_level,
            not_valid_before,
            not_valid_after,
            report_link,
            st_link,
            cert_link,
            manufacturer_web,
            protection_profiles,
            maintenances,
            None,
            None,
            None,
        )

    def set_local_paths(
        self,
        report_pdf_dir: Optional[Union[str, Path]],
        st_pdf_dir: Optional[Union[str, Path]],
        report_txt_dir: Optional[Union[str, Path]],
        st_txt_dir: Optional[Union[str, Path]],
    ) -> None:
        if report_pdf_dir is not None:
            self.state.report_pdf_path = Path(report_pdf_dir) / (self.dgst + ".pdf")
        if st_pdf_dir is not None:
            self.state.st_pdf_path = Path(st_pdf_dir) / (self.dgst + ".pdf")
        if report_txt_dir is not None:
            self.state.report_txt_path = Path(report_txt_dir) / (self.dgst + ".txt")
        if st_txt_dir is not None:
            self.state.st_txt_path = Path(st_txt_dir) / (self.dgst + ".txt")

    @staticmethod
    def download_pdf_report(cert: "CommonCriteriaCert") -> "CommonCriteriaCert":
        exit_code: Union[str, int]
        if not cert.report_link:
            exit_code = "No link"
        else:
            exit_code = helpers.download_file(cert.report_link, cert.state.report_pdf_path)
        if exit_code != requests.codes.ok:
            error_msg = f"failed to download report from {cert.report_link}, code: {exit_code}"
            logger.error(f"Cert dgst: {cert.dgst} " + error_msg)
            cert.state.report_download_ok = False
            cert.state.errors.append(error_msg)
        return cert

    @staticmethod
    def download_pdf_target(cert: "CommonCriteriaCert") -> "CommonCriteriaCert":
        exit_code: Union[str, int]
        if not cert.st_link:
            exit_code = "No link"
        else:
            exit_code = helpers.download_file(cert.st_link, cert.state.st_pdf_path)
        if exit_code != requests.codes.ok:
            error_msg = f"failed to download ST from {cert.report_link}, code: {exit_code}"
            logger.error(f"Cert dgst: {cert.dgst}" + error_msg)
            cert.state.st_download_ok = False
            cert.state.errors.append(error_msg)
        return cert

    @staticmethod
    def convert_report_pdf(cert: "CommonCriteriaCert") -> "CommonCriteriaCert":
        exit_code = helpers.convert_pdf_file(cert.state.report_pdf_path, cert.state.report_txt_path)
        if exit_code != constants.RETURNCODE_OK:
            error_msg = "failed to convert report pdf->txt"
            logger.error(f"Cert dgst: {cert.dgst}" + error_msg)
            cert.state.report_convert_ok = False
            cert.state.errors.append(error_msg)
        return cert

    @staticmethod
    def convert_target_pdf(cert: "CommonCriteriaCert") -> "CommonCriteriaCert":
        exit_code = helpers.convert_pdf_file(cert.state.st_pdf_path, cert.state.st_txt_path)
        if exit_code != constants.RETURNCODE_OK:
            error_msg = "failed to convert security target pdf->txt"
            logger.error(f"Cert dgst: {cert.dgst}" + error_msg)
            cert.state.st_convert_ok = False
            cert.state.errors.append(error_msg)
        return cert

    @staticmethod
    def extract_st_pdf_metadata(cert: "CommonCriteriaCert") -> "CommonCriteriaCert":
        response, cert.pdf_data.st_metadata = helpers.extract_pdf_metadata(cert.state.st_pdf_path)
        if response != constants.RETURNCODE_OK:
            cert.state.st_extract_ok = False
            cert.state.errors.append(response)
        return cert

    @staticmethod
    def extract_report_pdf_metadata(cert: "CommonCriteriaCert") -> "CommonCriteriaCert":
        response, cert.pdf_data.report_metadata = helpers.extract_pdf_metadata(cert.state.report_pdf_path)
        if response != constants.RETURNCODE_OK:
            cert.state.report_extract_ok = False
            cert.state.errors.append(response)
        return cert

    @staticmethod
    def extract_st_pdf_frontpage(cert: "CommonCriteriaCert") -> "CommonCriteriaCert":
        cert.pdf_data.st_frontpage = {}

        for header_type, associated_header_func in HEADERS.items():
            response, cert.pdf_data.st_frontpage[header_type] = associated_header_func(cert.state.st_txt_path)

            if response != constants.RETURNCODE_OK:
                cert.state.st_extract_ok = False
                if not cert.state.errors:
                    cert.state.errors = []
                cert.state.errors.append(response)

        return cert

    @staticmethod
    def extract_report_pdf_frontpage(cert: "CommonCriteriaCert") -> "CommonCriteriaCert":
        cert.pdf_data.report_frontpage = {}

        for header_type, associated_header_func in HEADERS.items():
            response, cert.pdf_data.report_frontpage[header_type] = associated_header_func(cert.state.report_txt_path)

            if response != constants.RETURNCODE_OK:
                cert.state.report_extract_ok = False
                if not cert.state.errors:
                    cert.state.errors = []
                cert.state.errors.append(response)

        return cert

    @staticmethod
    def extract_report_pdf_keywords(cert: "CommonCriteriaCert") -> "CommonCriteriaCert":
        response, cert.pdf_data.report_keywords = helpers.extract_keywords(cert.state.report_txt_path)
        if response != constants.RETURNCODE_OK:
            cert.state.report_extract_ok = False
        return cert

    @staticmethod
    def extract_st_pdf_keywords(cert: "CommonCriteriaCert") -> "CommonCriteriaCert":
        response, cert.pdf_data.st_keywords = helpers.extract_keywords(cert.state.st_txt_path)
        if response != constants.RETURNCODE_OK:
            cert.state.st_extract_ok = False
            cert.state.errors.append(response)
        return cert

    def _compute_heuristics_version(self) -> None:
        self.heuristics.extracted_versions = helpers.compute_heuristics_version(self.name)

    def compute_heuristics_cpe_match(self, cpe_classifier: CPEClassifier) -> None:
        self._compute_heuristics_version()
        assert self.heuristics.extracted_versions is not None

        self.heuristics.cpe_matches = cpe_classifier.predict_single_cert(
            self.manufacturer, self.name, self.heuristics.extracted_versions
        )

    def compute_heuristics_cert_lab(self) -> None:
        if not self.pdf_data:
            logger.error("Cannot compute sample lab when pdf files were not processed.")
            return
        self.heuristics.cert_lab = self.pdf_data.cert_lab

    def compute_heuristics_cert_id(self, all_cert_ids: Set[str]):
        if not self.pdf_data:
            logger.warning("Cannot compute sample id when pdf files were not processed.")
            return
        self.heuristics.cert_id = self.pdf_data.cert_id
        self.normalize_cert_id(all_cert_ids)

    @staticmethod
    def _is_anssi_cert(cert_id: str) -> bool:
        return cert_id.startswith("ANSS")

    @staticmethod
    def _fix_anssi_cert_id(cert_id: str) -> str:
        new_cert_id = cert_id

        if new_cert_id.startswith("ANSSi"):  # mistyped ANSSi
            new_cert_id = "ANSSI" + new_cert_id[4:]

        # Bug - getting out of index - ANSSI-2009/30
        # TMP solution
        if len(new_cert_id) >= len("ANSSI-CC-0000") + 1:
            if (
                new_cert_id[len("ANSSI-CC-0000")] == "_"
            ):  # _ instead of / after year (ANSSI-CC-2010_40 -> ANSSI-CC-2010/40)
                new_cert_id = new_cert_id[: len("ANSSI-CC-0000")] + "/" + new_cert_id[len("ANSSI-CC-0000") + 1 :]

        if "_" in new_cert_id:  # _ instead of -
            new_cert_id = new_cert_id.replace("_", "-")

        return new_cert_id

    @staticmethod
    def _is_bsi_cert(cert_id: str) -> bool:
        return cert_id.startswith("BSI-DSZ-CC-")

    @staticmethod
    def _extract_bsi_parts(bsi_parts: List[str]) -> Tuple:
        cert_num = None
        cert_version = None
        cert_year = None

        if len(bsi_parts) > 3:
            cert_num = bsi_parts[3]
        if len(bsi_parts) > 4:
            if bsi_parts[4].startswith("V") or bsi_parts[4].startswith("v"):
                cert_version = bsi_parts[4].upper()  # get version in uppercase
            else:
                cert_year = bsi_parts[4]
        if len(bsi_parts) > 5:
            cert_year = bsi_parts[5]

        return cert_num, cert_version, cert_year

    @staticmethod
    def _fix_bsi_cert_id(cert_id: str, all_cert_ids: Set[str]) -> str:
        start_year = 1996
        limit_year = datetime.now().year + 1
        bsi_parts = cert_id.split("-")

        cert_num, cert_version, cert_year = CommonCriteriaCert._extract_bsi_parts(bsi_parts)
        if cert_year is None:
            for year in range(start_year, limit_year):
                cert_id_possible = cert_id + "-" + str(year)

                if cert_id_possible in all_cert_ids:
                    # we found version with year
                    cert_year = str(year)
                    break

        # reconstruct BSI number again
        new_cert_id = "BSI-DSZ-CC"
        if cert_num is not None:
            new_cert_id += "-" + cert_num
        if cert_version is not None:
            new_cert_id += "-" + cert_version
        if cert_year is not None:
            new_cert_id += "-" + cert_year

        return new_cert_id

    @staticmethod
    def _is_spain_cert_id(cert_id: str) -> bool:
        return "-INF-" in cert_id

    @staticmethod
    def _fix_spain_cert_id(cert_id: str) -> str:
        spain_parts = cert_id.split("-")
        cert_year = spain_parts[0]
        cert_batch = spain_parts[1]
        cert_num = spain_parts[3]

        if "v" in cert_num:
            cert_num = cert_num[: cert_num.find("v")]
        if "V" in cert_num:
            cert_num = cert_num[: cert_num.find("V")]

        new_cert_id = f"{cert_year}-{cert_batch}-INF-{cert_num}"  # drop version

        return new_cert_id

    @staticmethod
    def _is_ocsi_cert_id(cert_id: str) -> bool:
        return "OCSI/CERT" in cert_id

    @staticmethod
    def _fix_ocsi_cert_id(cert_id: str) -> str:
        new_cert_id = cert_id
        if not new_cert_id.endswith("/RC"):
            new_cert_id = cert_id + "/RC"

        return new_cert_id

    def get_cert_laboratory(self) -> str:
        if not self.heuristics.cert_id:
            raise ValueError("Cert ID was None but cert laboratory was to be computed based on its value.")
        cert_id = self.heuristics.cert_id.strip()

        if CommonCriteriaCert._is_anssi_cert(cert_id):
            return "anssi"

        if CommonCriteriaCert._is_bsi_cert(cert_id):
            return "bsi"

        if CommonCriteriaCert._is_spain_cert_id(cert_id):
            return "spain"

        if CommonCriteriaCert._is_ocsi_cert_id(cert_id):
            return "ocsi"

        return "unknown"

    def normalize_cert_id(self, all_cert_ids: Set[str]) -> None:
        fix_methods: Dict[str, Callable] = {
            "anssi": CommonCriteriaCert._fix_anssi_cert_id,
            "bsi": partial(CommonCriteriaCert._fix_bsi_cert_id, all_cert_ids=all_cert_ids),
            "spain": CommonCriteriaCert._fix_spain_cert_id,
            "ocsi": CommonCriteriaCert._fix_ocsi_cert_id,
        }

        try:
            cert_lab = self.get_cert_laboratory()
        except ValueError:
            return None

        # No need for any fix, bcs we do not know how
        if cert_lab == "unknown":
            return None

        self.heuristics.cert_id = fix_methods[cert_lab](self.pdf_data.cert_id)



import shutil
import tempfile
from pathlib import Path
from typing import ClassVar, Dict, List
from unittest import TestCase

import tests.data.test_cc_heuristics
from sec_certs import constants
from sec_certs.dataset.common_criteria import CCDataset
from sec_certs.dataset.cpe import CPEDataset
from sec_certs.dataset.cve import CVEDataset
from sec_certs.sample.common_criteria import CommonCriteriaCert
from sec_certs.sample.cpe import CPE
from sec_certs.sample.cve import CVE
from sec_certs.sample.protection_profile import ProtectionProfile


class TestCommonCriteriaHeuristics(TestCase):
    dataset_json_path: ClassVar[Path] = Path(tests.data.test_cc_heuristics.__path__[0]) / "vulnerable_dataset.json"
    data_dir_path: ClassVar[Path] = dataset_json_path.parent
    tmp_dir: ClassVar[tempfile.TemporaryDirectory]
    cc_dset: CCDataset
    cve_dset: CVEDataset
    cves: List[CVE]
    cpe_dset: CPEDataset
    cpes: List[CPE]

    @classmethod
    def setUpClass(cls) -> None:
        cls.tmp_dir = tempfile.TemporaryDirectory()
        shutil.copytree(cls.data_dir_path, cls.tmp_dir.name, dirs_exist_ok=True)

        cls.cc_dset = CCDataset.from_json(Path(cls.tmp_dir.name) / "vulnerable_dataset.json")
        cls.cc_dset.process_protection_profiles()
        cls.cc_dset.download_all_pdfs()
        cls.cc_dset.convert_all_pdfs()
        cls.cc_dset._extract_data()
        cls.cc_dset._compute_heuristics(use_nist_cpe_matching_dict=False)

        cpe_single_sign_on = CPE(
            "cpe:2.3:a:ibm:security_access_manager_for_enterprise_single_sign-on:8.2.2:*:*:*:*:*:*:*",
            "IBM Security Access Manager For Enterprise Single Sign-On 8.2.2",
        )

        cls.cpes = [
            cpe_single_sign_on,
            CPE(
                "cpe:2.3:a:ibm:security_key_lifecycle_manager:2.6.0.1:*:*:*:*:*:*:*",
                "IBM Security Key Lifecycle Manager 2.6.0.1",
            ),
            CPE(
                "cpe:2.3:a:semperplugins:all_in_one_seo_pack:1.3.6.4:*:*:*:*:wordpress:*:*",
                "Semper Plugins All in One SEO Pack 1.3.6.4 for WordPress",
            ),
            CPE(
                "cpe:2.3:a:tracker-software:pdf-xchange_lite_printer:6.0.320.0:*:*:*:*:*:*:*",
                "Tracker Software PDF-XChange Lite Printer 6.0.320.0",
            ),
        ]
        cls.cpe_dset = CPEDataset(True, Path("../"), {x.uri: x for x in cls.cpes})

        cls.cves = [
            CVE(
                "CVE-2017-1732",
                [cpe_single_sign_on],
                CVE.Impact(5.3, "MEDIUM", 3.9, 1.4),
                "2021-05-26T04:15Z",
            ),
            CVE("CVE-2019-4513", [cpe_single_sign_on], CVE.Impact(8.2, "HIGH", 3.9, 4.2), "2000-05-26T04:15Z"),
        ]
        cls.cve_dset = CVEDataset({x.cve_id: x for x in cls.cves})
        cls.cve_dset.build_lookup_dict(use_nist_mapping=False)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmp_dir.cleanup()

    def test_load_cpe_dataset(self):
        json_cpe_dset = CPEDataset.from_json(self.data_dir_path / "auxillary_datasets" / "cpe_dataset.json")
        json_cpe_dset.json_path = Path("../")
        self.assertEqual(
            self.cpe_dset, json_cpe_dset, "CPE template dataset does not match CPE dataset loaded from json."
        )

    def test_cpe_lookup_dicts(self):
        self.assertEqual(
            self.cpe_dset.vendors,
            {"ibm", "tracker-software", "semperplugins"},
            "The set of versions in CPE dataset does not match template",
        )
        self.assertEqual(
            self.cpe_dset.vendor_to_versions,
            {"ibm": {"8.2.2", "2.6.0.1"}, "semperplugins": {"1.3.6.4"}, "tracker-software": {"6.0.320.0"}},
            "The CPE lookup dictionary vendor->version of CPE dataset does not match template.",
        )
        self.assertEqual(
            self.cpe_dset.vendor_version_to_cpe,
            {
                ("ibm", "8.2.2"): {
                    CPE(
                        "cpe:2.3:a:ibm:security_access_manager_for_enterprise_single_sign-on:8.2.2:*:*:*:*:*:*:*",
                        "IBM Security Access Manager For Enterprise Single Sign-On 8.2.2",
                    )
                },
                ("ibm", "2.6.0.1"): {
                    CPE(
                        "cpe:2.3:a:ibm:security_key_lifecycle_manager:2.6.0.1:*:*:*:*:*:*:*",
                        "IBM Security Key Lifecycle Manager 2.6.0.1",
                    )
                },
                ("semperplugins", "1.3.6.4"): {
                    CPE(
                        "cpe:2.3:a:semperplugins:all_in_one_seo_pack:1.3.6.4:*:*:*:*:wordpress:*:*",
                        "Semper Plugins All in One SEO Pack 1.3.6.4 for WordPress",
                    )
                },
                ("tracker-software", "6.0.320.0"): {
                    CPE(
                        "cpe:2.3:a:tracker-software:pdf-xchange_lite_printer:6.0.320.0:*:*:*:*:*:*:*",
                        "Tracker Software PDF-XChange Lite Printer 6.0.320.0",
                    )
                },
            },
            "The CPE lookup dictionary (vendor,version)->cpe does not match the template.",
        )

    def test_cpe_parsing(self):
        potentially_problematic_cpes = {
            'cpe:2.3:a:bayashi:dopvstar\::0091:*:*:*:*:*:*:*"': ("bayashi", "dopvstar:", "0091"),  # noqa: W605
            "cpe:2.3:a:moundlabs:\:\:mound\:\::2.1.6:*:*:*:*:*:*:*": ("moundlabs", "::mound::", "2.1.6"),  # noqa: W605
            "cpe:2.3:a:lemonldap-ng:lemonldap\:\::*:*:*:*:*:*:*:*": (  # noqa: W605
                "lemonldap-ng",
                "lemonldap::",
                constants.CPE_VERSION_NA,
            ),
            "cpe:2.3:o:cisco:nx-os:5.0\\\\\\(3\\\\\\)u5\\\\\\(1g\\\\\\):*:*:*:*:*:*:*": (
                "cisco",
                "nx-os",
                "5.0\\(3\\)u5\\(1g\\)",
            ),
            "cpe:2.3:a:\\@thi.ng\\/egf_project:\\@thi.ng\\/egf:-:*:*:*:*:node.js:*:*": (
                "@thi.ng/egf project",
                "@thi.ng/egf",
                "-",
            ),
            "cpe:2.3:a:oracle:communications_diameter_signaling_router_idih\\:::*:*:*:*:*:*:*": (
                "oracle",
                "communications diameter signaling router idih:",
                constants.CPE_VERSION_NA,
            ),
        }

        for uri, tpl in potentially_problematic_cpes.items():
            cpe = CPE(uri)
            self.assertEqual(cpe.vendor, tpl[0], "Parsed CPE vendor differs from expected vendor. Broken escaping?")
            self.assertEqual(
                cpe.item_name, tpl[1], "Parsed CPE item name differs from expected item name. Broken escaping?"
            )
            self.assertEqual(cpe.version, tpl[2], "Parsed CPE version differs from expected version. Broken escaping?")

    def test_cve_lookup_dicts(self):
        alt_lookup = {x: set(y) for x, y in self.cve_dset.cpe_to_cve_ids_lookup.items()}
        self.assertEqual(
            alt_lookup,
            {
                "cpe:2.3:a:ibm:security_access_manager_for_enterprise_single_sign-on:8.2.2:*:*:*:*:*:*:*": {
                    x.cve_id for x in self.cves
                }
            },
            "The CVE lookup dicionary cve-> affected cpes does not match the template",
        )

    def test_load_cve_dataset(self):
        json_cve_dset = CVEDataset.from_json(self.data_dir_path / "auxillary_datasets" / "cve_dataset.json")
        self.assertEqual(
            self.cve_dset, json_cve_dset, "CVE template dataset does not match CVE dataset loaded from json."
        )

    def test_match_cpe(self):
        self.assertTrue(
            self.cpes[0].uri in self.cc_dset["ebd276cca70fd723"].heuristics.cpe_matches,
            "The CPE matching algorithm did not find the right CPE.",
        )
        self.assertTrue(
            len(self.cc_dset["ebd276cca70fd723"].heuristics.cpe_matches) == 1, "Exactly one CPE match should be found."
        )

    def test_find_related_cves(self):
        self.cc_dset["ebd276cca70fd723"].heuristics.cpe_matches = [self.cpes[0].uri]
        self.cc_dset.compute_related_cves(use_nist_cpe_matching_dict=False)
        self.assertEqual(
            {x.cve_id for x in self.cves},
            self.cc_dset["ebd276cca70fd723"].heuristics.related_cves,
            "The computed CVEs do not match the excpected CVEs",
        )

    def test_version_extraction(self):
        self.assertEqual(
            self.cc_dset["ebd276cca70fd723"].heuristics.extracted_versions,
            {"8.2"},
            "The version extracted from the sample does not match the template",
        )
        new_cert = CommonCriteriaCert(
            "",
            "",
            "IDOneClassIC Card : ID-One Cosmo 64 RSA v5.4 and applet IDOneClassIC v1.0 embedded on P5CT072VOP",
            "",
            "",
            "",
            None,
            None,
            "",
            "",
            "",
            "",
            set(),
            set(),
            None,
            None,
            None,
        )
        new_cert._compute_heuristics_version()
        self.assertEqual(
            set(new_cert.heuristics.extracted_versions),
            {"5.4", "1.0"},
            "The extracted versions do not match the template.",
        )

    def test_cert_lab_heuristics(self):
        self.assertEqual(self.cc_dset["ebd276cca70fd723"].heuristics.cert_lab, ["BSI"])

    def test_cert_id_heuristics(self):
        self.assertEqual(self.cc_dset["ebd276cca70fd723"].heuristics.cert_id, "BSI-DSZ-CC-0683-2014")

    def test_keywords_heuristics(self):
        extracted_keywords: Dict = self.cc_dset["ebd276cca70fd723"].pdf_data.st_keywords

        self.assertTrue("rules_security_level" in extracted_keywords)
        self.assertEqual(extracted_keywords["rules_security_level"]["EAL3"], 1)

        self.assertTrue("rules_security_assurance_components" in extracted_keywords)
        self.assertEqual(extracted_keywords["rules_security_assurance_components"]["ADV_ARC.1"], 1)
        self.assertEqual(extracted_keywords["rules_security_assurance_components"]["ADV_FSP.3"], 1)
        self.assertEqual(extracted_keywords["rules_security_assurance_components"]["ADV_TDS.2"], 1)

        self.assertTrue("rules_crypto_algs" in extracted_keywords)
        self.assertEqual(extracted_keywords["rules_crypto_algs"]["AES"], 2)

        self.assertTrue("rules_block_cipher_modes" in extracted_keywords)
        self.assertEqual(extracted_keywords["rules_block_cipher_modes"]["CBC"], 2)

    def test_protection_profiles_matching(self):
        artificial_pp: ProtectionProfile = ProtectionProfile(
            "Korean National Protection Profile for Single Sign On V1.0",
            "http://www.commoncriteriaportal.org/files/ppfiles/KECS-PP-0822-2017%20Korean%20National%20PP%20for%20Single%20Sign%20On%20V1.0(eng).pdf",
        )
        self.cc_dset["ebd276cca70fd723"].protection_profiles = {artificial_pp}
        expected_pp: ProtectionProfile = ProtectionProfile(
            "Korean National Protection Profile for Single Sign On V1.0",
            "http://www.commoncriteriaportal.org/files/ppfiles/KECS-PP-0822-2017%20Korean%20National%20PP%20for%20Single%20Sign%20On%20V1.0(eng).pdf",
            frozenset(["KECS-PP-0822-2017 SSO V1.0"]),
        )
        self.cc_dset.process_protection_profiles(to_download=False)
        self.assertSetEqual(self.cc_dset["ebd276cca70fd723"].protection_profiles, {expected_pp})

    def test_single_record_dependency_heuristics(self):
        # Single record in daset is not affecting nor affected by other records
        heuristics = self.cc_dset["ebd276cca70fd723"].heuristics
        self.assertEqual(heuristics.report_references.directly_referenced_by, None)
        self.assertEqual(heuristics.report_references.indirectly_referenced_by, None)
        self.assertEqual(heuristics.report_references.directly_referencing, None)
        self.assertEqual(heuristics.report_references.indirectly_referencing, None)

    def test_dependency_dataset(self):
        dependency_dataset = CCDataset.from_json(self.data_dir_path / "dependency_dataset.json")
        dependency_dataset._compute_dependencies()
        test_cert = dependency_dataset["692e91451741ef49"]

        self.assertEqual(test_cert.heuristics.report_references.directly_referenced_by, {"BSI-DSZ-CC-0370-2006"})
        self.assertEqual(
            test_cert.heuristics.report_references.indirectly_referenced_by,
            {"BSI-DSZ-CC-0370-2006", "BSI-DSZ-CC-0517-2009"},
        )
        self.assertEqual(test_cert.heuristics.report_references.directly_referencing, {"BSI-DSZ-CC-0268-2005"})
        self.assertEqual(test_cert.heuristics.report_references.indirectly_referencing, {"BSI-DSZ-CC-0268-2005"})



from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Iterator, List, Mapping, Union

from sec_certs.dataset.dataset import logger
from sec_certs.helpers import tqdm
from sec_certs.sample.fips_mip import MIPSnapshot
from sec_certs.serialization.json import ComplexSerializableType


@dataclass
class MIPDataset(ComplexSerializableType):
    snapshots: List[MIPSnapshot]

    def __iter__(self) -> Iterator[MIPSnapshot]:
        yield from self.snapshots

    def __getitem__(self, item: int) -> MIPSnapshot:
        return self.snapshots.__getitem__(item)

    def __len__(self) -> int:
        return len(self.snapshots)

    @classmethod
    def from_dumps(cls, dump_path: Union[str, Path]) -> "MIPDataset":
        directory = Path(dump_path)
        fnames = list(directory.glob("*"))
        snapshots = []
        for dump_path in tqdm(sorted(fnames), total=len(fnames)):
            try:
                snapshots.append(MIPSnapshot.from_dump(dump_path))
            except Exception as e:
                logger.error(e)
        return cls(snapshots)

    def to_dict(self) -> Dict[str, List[MIPSnapshot]]:
        return {"snapshots": list(self.snapshots)}

    @classmethod
    def from_dict(cls, dct: Mapping) -> "MIPDataset":
        return cls(dct["snapshots"])



import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, Final, List
from unittest import TestCase

import tests.data.test_fips_oop
from sec_certs.config.configuration import config
from sec_certs.dataset.fips import FIPSDataset
from sec_certs.dataset.fips_algorithm import FIPSAlgorithmDataset
from sec_certs.helpers import fips_dgst
from tests.fips_test_utils import generate_html


def _set_up_dataset(td, certs):
    dataset = FIPSDataset({}, Path(td), "test_dataset", "fips_test_dataset")
    generate_html(certs, td + "/test_search.html")
    dataset.get_certs_from_web(test=td + "/test_search.html", no_download_algorithms=True)
    return dataset


def _set_up_dataset_for_full(td, certs, cpe_dset_path: Path, cve_dset_path: Path):
    dataset = _set_up_dataset(td, certs)

    dataset.auxillary_datasets_dir.mkdir(exist_ok=True)
    shutil.copyfile(cpe_dset_path, dataset.cpe_dataset_path)
    shutil.copyfile(cve_dset_path, dataset.cve_dataset_path)

    dataset.web_scan(set(certs))
    dataset.download_all_pdfs(set(certs))
    dataset.convert_all_pdfs()
    dataset.pdf_scan()
    dataset.extract_certs_from_tables(high_precision=True)
    dataset.algorithms = FIPSAlgorithmDataset.from_json(Path(__file__).parent / "data/test_fips_oop/algorithms.json")
    dataset.finalize_results(use_nist_cpe_matching_dict=False, perform_cpe_heuristics=False)
    return dataset


class TestFipsOOP(TestCase):
    data_dir: Final[Path] = Path(tests.data.test_fips_oop.__path__[0])
    cpe_dset_path: Final[Path] = data_dir.parent / "test_cc_heuristics/auxillary_datasets/cpe_dataset.json"
    cve_dset_path: Final[Path] = data_dir.parent / "test_cc_heuristics/auxillary_datasets/cve_dataset.json"
    certs_to_parse: Final[Dict[str, List[str]]] = {
        "microsoft": [
            "3095",
            "3651",
            "3093",
            "3090",
            "3197",
            "3196",
            "3089",
            "3195",
            "3480",
            "3615",
            "3194",
            "3091",
            "3690",
            "3644",
            "3527",
            "3094",
            "3544",
            "3096",
            "3092",
        ],
        "redhat": [
            "2630",
            "2721",
            "2997",
            "2441",
            "2711",
            "2633",
            "2798",
            "3613",
            "3733",
            "2908",
            "2446",
            "2742",
            "2447",
        ],
        "docusign": ["3850", "2779", "2860", "2665", "1883", "3518", "3141", "2590"],
        "referencing_openssl": [
            "3493",
            "3495",
            "3711",
            "3176",
            "3488",
            "3126",
            "3269",
            "3524",
            "3220",
            "2398",
            "3543",
            "2676",
            "3313",
            "3363",
            "3608",
            "3158",
        ],
    }

    @classmethod
    def setUpClass(cls) -> None:
        config.load(cls.data_dir.parent / "settings_test.yaml")

    def test_regress_125(self):
        with TemporaryDirectory() as tmp_dir:
            dst = _set_up_dataset(tmp_dir, ["3493", "3495"])
            self.assertEqual(set(dst.certs), {fips_dgst("3493"), fips_dgst("3495")})
            self.assertIsInstance(dst.certs[fips_dgst("3493")].cert_id, int)
            self.assertEqual(dst.certs[fips_dgst("3493")].cert_id, 3493)

    def test_size(self):
        for certs in self.certs_to_parse.values():
            with TemporaryDirectory() as tmp_dir:
                dataset = _set_up_dataset(tmp_dir, certs)
                self.assertEqual(len(dataset.certs), len(certs), "Wrong number of parsed certs")

    def test_connections_microsoft(self):
        certs = self.certs_to_parse["microsoft"]
        with TemporaryDirectory() as tmp_dir:
            dataset = _set_up_dataset_for_full(tmp_dir, certs, self.cpe_dset_path, self.cve_dset_path)

            self.assertEqual(
                set(dataset.certs[fips_dgst("3095")].heuristics.st_references.directly_referencing), {"3096"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3095")].heuristics.web_references.directly_referencing),
                {"3093", "3096", "3094"},
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3651")].heuristics.st_references.directly_referencing), {"3615"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3093")].heuristics.st_references.directly_referencing), {"3091"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3093")].heuristics.web_references.directly_referencing), {"3090", "3091"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3090")].heuristics.st_references.directly_referencing), {"3089"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3197")].heuristics.web_references.directly_referencing),
                {"3195", "3096", "3196", "3644", "3651"},
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3196")].heuristics.st_references.directly_referencing), {"3091"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3196")].heuristics.web_references.directly_referencing),
                {"3194", "3091", "3480", "3615"},
            )
            self.assertIsNone(dataset.certs[fips_dgst("3089")].heuristics.st_references.directly_referencing)
            self.assertEqual(
                set(dataset.certs[fips_dgst("3195")].heuristics.st_references.directly_referencing), {"3091"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3195")].heuristics.web_references.directly_referencing),
                {"3194", "3091", "3480"},
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3480")].heuristics.st_references.directly_referencing), {"3089"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3615")].heuristics.st_references.directly_referencing), {"3089"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3194")].heuristics.st_references.directly_referencing), {"3089"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3091")].heuristics.st_references.directly_referencing), {"3089"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3690")].heuristics.st_references.directly_referencing), {"3651"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3690")].heuristics.web_references.directly_referencing),
                {"3644", "3196", "3651"},
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3644")].heuristics.st_references.directly_referencing), {"3615"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3527")].heuristics.st_references.directly_referencing), {"3091"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3527")].heuristics.web_references.directly_referencing), {"3090", "3091"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3094")].heuristics.st_references.directly_referencing), {"3091"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3544")].heuristics.st_references.directly_referencing), {"3096"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3544")].heuristics.web_references.directly_referencing),
                {"3093", "3096", "3527"},
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3096")].heuristics.st_references.directly_referencing), {"3091"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3096")].heuristics.web_references.directly_referencing),
                {"3090", "3194", "3091", "3480"},
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3092")].heuristics.web_references.directly_referencing),
                {"3093", "3195", "3096", "3644", "3651"},
            )

    def test_connections_redhat(self):
        certs = self.certs_to_parse["redhat"]
        with TemporaryDirectory() as tmp_dir:
            dataset = _set_up_dataset_for_full(tmp_dir, certs, self.cpe_dset_path, self.cve_dset_path)
            self.assertEqual(
                set(dataset.certs[fips_dgst("2630")].heuristics.st_references.directly_referencing), {"2441"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2633")].heuristics.st_references.directly_referencing), {"2441"}
            )
            self.assertIsNone(dataset.certs[fips_dgst("2441")].heuristics.st_references.directly_referencing)
            self.assertEqual(
                set(dataset.certs[fips_dgst("2997")].heuristics.st_references.directly_referencing), {"2711"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2446")].heuristics.st_references.directly_referencing), {"2441"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2447")].heuristics.st_references.directly_referencing), {"2441"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3733")].heuristics.st_references.directly_referencing), {"2441"}
            )
            self.assertIsNone(dataset.certs[fips_dgst("2441")].heuristics.st_references.directly_referencing)
            self.assertIsNone(dataset.certs[fips_dgst("2711")].heuristics.st_references.directly_referencing)
            self.assertEqual(
                set(dataset.certs[fips_dgst("2908")].heuristics.st_references.directly_referencing), {"2711"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3613")].heuristics.st_references.directly_referencing), {"2997"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2721")].heuristics.st_references.directly_referencing), {"2441"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2721")].heuristics.web_references.directly_referencing), {"2441", "2711"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2798")].heuristics.st_references.directly_referencing), {"2721"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2798")].heuristics.web_references.directly_referencing), {"2711", "2721"}
            )
            self.assertIsNone(dataset.certs[fips_dgst("2711")].heuristics.st_references.directly_referencing)
            self.assertEqual(
                set(dataset.certs[fips_dgst("2997")].heuristics.st_references.directly_referencing), {"2711"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2742")].heuristics.st_references.directly_referencing), {"2721"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2742")].heuristics.web_references.directly_referencing), {"2721", "2711"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2721")].heuristics.st_references.directly_referencing), {"2441"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2721")].heuristics.web_references.directly_referencing), {"2441", "2711"}
            )

    def test_docusign_chunk(self):
        certs = self.certs_to_parse["docusign"]
        with TemporaryDirectory() as tmp_dir:
            dataset = _set_up_dataset_for_full(tmp_dir, certs, self.cpe_dset_path, self.cve_dset_path)
            self.assertEqual(
                set(dataset.certs[fips_dgst("3850")].heuristics.st_references.directly_referencing), {"1883"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3850")].heuristics.web_references.directly_referencing), {"1883"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2779")].heuristics.st_references.directly_referencing), {"1883"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2860")].heuristics.st_references.directly_referencing), {"1883"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2665")].heuristics.st_references.directly_referencing), {"1883"}
            )
            self.assertIsNone(dataset.certs[fips_dgst("1883")].heuristics.st_references.directly_referencing)
            self.assertEqual(
                set(dataset.certs[fips_dgst("3518")].heuristics.st_references.directly_referencing), {"1883"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3141")].heuristics.st_references.directly_referencing), {"1883"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2590")].heuristics.st_references.directly_referencing), {"1883"}
            )

    def test_openssl_chunk(self):
        certs = self.certs_to_parse["referencing_openssl"]
        with TemporaryDirectory() as tmp_dir:
            dataset = _set_up_dataset_for_full(tmp_dir, certs, self.cpe_dset_path, self.cve_dset_path)
            self.assertEqual(
                set(dataset.certs[fips_dgst("3493")].heuristics.st_references.directly_referencing), {"2398"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3495")].heuristics.st_references.directly_referencing), {"2398"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3711")].heuristics.st_references.directly_referencing), {"3220"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3176")].heuristics.st_references.directly_referencing), {"2398"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3488")].heuristics.st_references.directly_referencing), {"2398"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3126")].heuristics.st_references.directly_referencing), {"2398"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3126")].heuristics.st_references.directly_referencing), {"2398"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3126")].heuristics.web_references.directly_referencing), {"2398"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3269")].heuristics.st_references.directly_referencing), {"3220"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3524")].heuristics.st_references.directly_referencing), {"3220"}
            )
            self.assertIsNone(dataset.certs[fips_dgst("3220")].heuristics.st_references.directly_referencing)
            self.assertIsNone(dataset.certs[fips_dgst("3220")].heuristics.web_references.directly_referencing)
            self.assertIsNone(dataset.certs[fips_dgst("2398")].heuristics.st_references.directly_referencing)
            self.assertEqual(
                set(dataset.certs[fips_dgst("3543")].heuristics.st_references.directly_referencing), {"2398"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("2676")].heuristics.st_references.directly_referencing), {"2398"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3313")].heuristics.st_references.directly_referencing), {"3220"}
            )
            self.assertIsNone(dataset.certs[fips_dgst("3363")].heuristics.st_references.directly_referencing)
            self.assertEqual(
                set(dataset.certs[fips_dgst("3608")].heuristics.st_references.directly_referencing), {"2398"}
            )
            self.assertEqual(
                set(dataset.certs[fips_dgst("3158")].heuristics.st_references.directly_referencing), {"2398"}
            )



import itertools
import logging
import operator
import re
from typing import Dict, List, Optional, Set, Tuple

from rapidfuzz import fuzz
from sklearn.base import BaseEstimator

import sec_certs.helpers as helpers
from sec_certs import constants
from sec_certs.sample.cpe import CPE

logger = logging.getLogger(__name__)


class CPEClassifier(BaseEstimator):
    """
    Class that can predict CPE matches for certificate instances.
    Adheres to sklearn BaseEstimator interface.
    Fit method is called on list of CPEs and build two look-up dictionaries, see description of attributes.
    """

    vendor_to_versions_: Dict[str, Set[str]]  # Key: CPE vendor, Value: versions of all CPE records of that vendor
    vendor_version_to_cpe_: Dict[
        Tuple[str, str], Set[CPE]
    ]  # Key: (CPE vendor, version), Value: All CPEs that are of (vendor, version)
    vendors_: Set[str]

    def __init__(self, match_threshold: int = 80, n_max_matches: int = 10):
        self.match_threshold = match_threshold
        self.n_max_matches = n_max_matches

    def fit(self, X: List[CPE], y: Optional[List[str]] = None) -> "CPEClassifier":
        """
        Just creates look-up structures from provided list of CPEs
        @param X: List of CPEs that can be matched with predict()
        @param y: will be ignored, specified to adhere to sklearn BaseEstimator interface
        """
        self.build_lookup_structures(X)
        return self

    @staticmethod
    def filter_short_cpes(cpes: List[CPE]) -> List[CPE]:
        """
        Short CPE items are super easy to match with 100% rank, but they are hardly informative. This method discards them.
        @param cpes: List of CPEs to filtered
        @return All CPEs in cpes variable which item name has at least 4 characters.
        """
        return list(filter(lambda x: x.item_name is not None and len(x.item_name) > 3, cpes))

    def build_lookup_structures(self, X: List[CPE]) -> None:
        """
        Builds several look-up dictionaries for fast matching.
        - vendor_to_version_: each vendor is mapped to set of versions that appear in combination with vendor in CPE dataset
        - vendor_version_to_cpe_: Each (vendor, version) tuple is mapped to a set of CPE items that appear in combination with this tuple in CPE dataset
        - vendors_: Just aggregates set of vendors, used for prunning later on.
        @param X: List of CPEs that will be used to build the dictionaries
        """
        sufficiently_long_cpes = self.filter_short_cpes(X)
        self.vendor_to_versions_ = {x.vendor: set() for x in sufficiently_long_cpes}
        self.vendors_ = set(self.vendor_to_versions_.keys())
        self.vendor_version_to_cpe_ = dict()

        for cpe in helpers.tqdm(sufficiently_long_cpes, desc="Fitting the CPE classifier"):
            self.vendor_to_versions_[cpe.vendor].add(cpe.version)
            if (cpe.vendor, cpe.version) not in self.vendor_version_to_cpe_:
                self.vendor_version_to_cpe_[(cpe.vendor, cpe.version)] = {cpe}
            else:
                self.vendor_version_to_cpe_[(cpe.vendor, cpe.version)].add(cpe)

    def predict(self, X: List[Tuple[str, str, str]]) -> List[Optional[Set[str]]]:
        """
        Will predict CPE uris for List of Tuples (vendor, product name, identified versions in product name)
        @param X: tuples (vendor, product name, identified versions in product name)
        @return: List of CPE uris that correspond to given input, None if nothing was found.
        """
        return [self.predict_single_cert(x[0], x[1], x[2]) for x in helpers.tqdm(X, desc="Predicting")]

    def predict_single_cert(
        self,
        vendor: Optional[str],
        product_name: str,
        versions: Set[str],
        relax_version: bool = False,
        relax_title: bool = False,
    ) -> Optional[Set[str]]:
        """
        Predict List of CPE uris for triplet (vendor, product_name, list_of_version). The prediction is made as follows:
        1. Sanitize all strings
        2. Find vendors in CPE dataset that are related to the certificate
        3. Based on (vendors, versions) find all CPE items that are considered as candidates for match
        4. Compute string similarity of the candidate CPE matches and certificate name
        5. Evaluate best string similarity, if above threshold, declare it a match.
        6. If no CPE item is matched, we tried again but relax version and check CPEs that don't have their version specified.
        Also, we search for 100% CPE matches on item name instead of title.
        @param vendor: manufacturer of the certificate
        @param product_name: name of the certificate
        @param versions: List of versions that appear in the certificate name
        @param relax_version: bool, see step 6 above.
        @param relax_title: bool
        @return:
        """
        sanitized_vendor = CPEClassifier._discard_trademark_symbols(vendor).lower() if vendor else vendor
        sanitized_product_name = CPEClassifier._fully_sanitize_string(product_name) if product_name else product_name
        candidate_vendors = self.get_candidate_list_of_vendors(sanitized_vendor)
        candidates = self.get_candidate_cpe_matches(candidate_vendors, versions)
        ratings = [
            self.compute_best_match(cpe, sanitized_product_name, candidate_vendors, versions, relax_title=relax_title)
            for cpe in candidates
        ]
        threshold = self.match_threshold if not relax_version else 100
        final_matches_aux: List[Tuple[float, CPE]] = list(filter(lambda x: x[0] >= threshold, zip(ratings, candidates)))
        final_matches_aux = sorted(final_matches_aux, key=operator.itemgetter(0, 1), reverse=True)
        final_matches: Optional[Set[str]] = set(
            [x[1].uri for x in final_matches_aux[: self.n_max_matches] if x[1].uri is not None]
        )

        if not relax_title and not final_matches:
            final_matches = self.predict_single_cert(
                vendor, product_name, versions, relax_version=relax_version, relax_title=True
            )

        if not relax_version and not final_matches:
            final_matches = self.predict_single_cert(
                vendor, product_name, {constants.CPE_VERSION_NA}, relax_version=True, relax_title=relax_title
            )

        return final_matches if final_matches else None

    def compute_best_match(
        self,
        cpe: CPE,
        product_name: str,
        candidate_vendors: Optional[Set[str]],
        versions: Set[str],
        relax_title: bool = False,
    ) -> float:
        """
        Tries several different settings in which string similarity between CPE and certificate name is tested.
        For definition of string similarity, see rapidfuzz package on GitHub. Both token set ratio and partial ratio are tested,
        always both on CPE title and CPE item name.
        @param cpe: CPE to test
        @param product_name: name of the certificate
        @param candidate_vendors: vendors that appear in the certificate
        @param versions: versions that appear in the certificate
        @return: Maximal value of the four string similarities discussed above.
        """
        if relax_title:
            sanitized_title = (
                CPEClassifier._fully_sanitize_string(cpe.title)
                if cpe.title
                else CPEClassifier._fully_sanitize_string(
                    cpe.vendor + " " + cpe.item_name + " " + cpe.version + " " + cpe.update + " " + cpe.target_hw
                )
            )
        else:
            if cpe.title:
                sanitized_title = CPEClassifier._fully_sanitize_string(cpe.title)
            else:
                return 0
        sanitized_item_name = CPEClassifier._fully_sanitize_string(cpe.item_name)
        cert_stripped = CPEClassifier._strip_manufacturer_and_version(product_name, candidate_vendors, versions)

        token_set_ratio_on_title = fuzz.token_set_ratio(product_name, sanitized_title)
        partial_ratio_on_title = fuzz.partial_ratio(product_name, sanitized_title)
        ratings = [token_set_ratio_on_title, partial_ratio_on_title]

        if relax_title:
            token_set_ratio_on_item_name = fuzz.token_set_ratio(cert_stripped, sanitized_item_name)
            partial_ratio_on_item_name = fuzz.partial_ratio(cert_stripped, sanitized_item_name)
            ratings += [token_set_ratio_on_item_name, partial_ratio_on_item_name]

        return max(ratings)

    @staticmethod
    def _fully_sanitize_string(string: str) -> str:
        return CPEClassifier._replace_special_chars_with_space(CPEClassifier._discard_trademark_symbols(string.lower()))

    @staticmethod
    def _replace_special_chars_with_space(string: str) -> str:
        return re.sub(r"[^a-zA-Z0-9 \n\.]", " ", string)

    @staticmethod
    def _discard_trademark_symbols(string: str) -> str:
        return string.replace("®", "").replace("™", "")

    @staticmethod
    def _strip_manufacturer_and_version(string: str, manufacturers: Optional[Set[str]], versions: Set[str]) -> str:
        to_strip = versions | manufacturers if manufacturers else versions
        for x in to_strip:
            string = string.lower().replace(CPEClassifier._replace_special_chars_with_space(x.lower()), "").strip()
        return string

    def _process_manufacturer(self, manufacturer: str, result: Set) -> Set[str]:
        tokenized = manufacturer.split()
        if tokenized[0] in self.vendors_:
            result.add(tokenized[0])
        if len(tokenized) > 1 and tokenized[0] + tokenized[1] in self.vendors_:
            result.add(tokenized[0] + tokenized[1])

        # Below are completely manual fixes
        if "hewlett" in tokenized or "hewlett-packard" in tokenized or manufacturer == "hewlett packard":
            result.add("hp")
        if "thales" in tokenized:
            result.add("thalesesecurity")
            result.add("thalesgroup")
        if "stmicroelectronics" in tokenized:
            result.add("st")
        if "athena" in tokenized and "smartcard" in tokenized:
            result.add("athena-scs")
        if tokenized[0] == "the" and not result:
            candidate_result = self.get_candidate_list_of_vendors(" ".join(tokenized[1:]))
            return set(candidate_result) if candidate_result else set()

        return set(result) if result else set()

    def get_candidate_list_of_vendors(self, manufacturer: Optional[str]) -> Set[str]:
        """
        Given manufacturer name, this method will find list of plausible vendors from CPE dataset that are likely related.
        @param manufacturer: manufacturer
        @return: List of related manufacturers, None if nothing relevant is found.
        """
        result: Set[str] = set()
        if not manufacturer:
            return result

        splits = re.compile(r"[,/]").findall(manufacturer)

        if splits:
            vendor_tokens = set(
                itertools.chain.from_iterable([[x.strip() for x in manufacturer.split(s)] for s in splits])
            )
            result_aux = [self.get_candidate_list_of_vendors(x) for x in vendor_tokens]
            result_used = set(set(itertools.chain.from_iterable([x for x in result_aux if x])))
            return result_used if result_used else set()

        if manufacturer in self.vendors_:
            result.add(manufacturer)

        return self._process_manufacturer(manufacturer, result)

    def get_candidate_vendor_version_pairs(
        self, cert_candidate_cpe_vendors: Set[str], cert_candidate_versions: Set[str]
    ) -> Optional[List[Tuple[str, str]]]:
        """
        Given parameters, will return Pairs (cpe_vendor, cpe_version) that are relevant to a given sample
        @param cert_candidate_cpe_vendors: list of CPE vendors relevant to a sample
        @param cert_candidate_versions: List of versions heuristically extracted from the sample name
        @return: List of tuples (cpe_vendor, cpe_version) that can be used in the lookup table to search the CPE dataset.
        """

        def is_cpe_version_among_cert_versions(cpe_version: Optional[str], cert_versions: Set[str]) -> bool:
            def simple_startswith(seeked_version: str, checked_string: str) -> bool:
                if seeked_version == checked_string:
                    return True
                else:
                    return (
                        checked_string.startswith(seeked_version) and not checked_string[len(seeked_version)].isdigit()
                    )

            if not cpe_version:
                return False
            just_numbers = r"(\d{1,5})(\.\d{1,5})"  # TODO: The use of this should be double-checked
            for v in cert_versions:
                if (simple_startswith(v, cpe_version) and re.search(just_numbers, cpe_version)) or simple_startswith(
                    cpe_version, v
                ):
                    return True
            return False

        if not cert_candidate_cpe_vendors:
            return None

        candidate_vendor_version_pairs: List[Tuple[str, str]] = []
        for vendor in cert_candidate_cpe_vendors:
            viable_cpe_versions = self.vendor_to_versions_.get(vendor, set())
            matched_cpe_versions = [
                x for x in viable_cpe_versions if is_cpe_version_among_cert_versions(x, cert_candidate_versions)
            ]
            candidate_vendor_version_pairs.extend([(vendor, x) for x in matched_cpe_versions])
        return candidate_vendor_version_pairs

    def get_candidate_cpe_matches(self, candidate_vendors: Set[str], candidate_versions: Set[str]) -> List[CPE]:
        """
        Given List of candidate vendors and candidate versions found in certificate, candidate CPE matches are found
        @param candidate_vendors: List of version strings that were found in the certificate
        @param candidate_versions: List of vendor strings that were found in the certificate
        @return:
        """
        candidate_vendor_version_pairs = self.get_candidate_vendor_version_pairs(candidate_vendors, candidate_versions)
        return (
            list(
                itertools.chain.from_iterable([self.vendor_version_to_cpe_[x] for x in candidate_vendor_version_pairs])
            )
            if candidate_vendor_version_pairs
            else []
        )



#!/usr/bin/env python3
from setuptools import find_packages, setup

setup(
    name="sec-certs",
    author="Petr Svenda, Stanislav Bobon, Jan Jancar, Adam Janovsky, Jiri Michalik",
    author_email="svenda@fi.muni.cz",
    version_config=True,
    setup_requires=["setuptools-git-versioning"],
    packages=find_packages(),
    license="MIT",
    description="Tool for analysis of security certificates",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Topic :: Security",
        "Topic :: Security :: Cryptography",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
    ],
    python_requires=">=3.8",
    install_requires=open("requirements/requirements.in", "r").read().splitlines(),
    extras_require={
        "dev": open("requirements/dev_requirements.in", "r").read().splitlines(),
        "test": open("requirements/test_requirements.in", "r").read().splitlines(),
    },
    include_package_data=True,
    package_data={"sec_certs": ["settings.yaml", "settings-schema.json"]},
    entry_points={"console_scripts": ["cc-certs=cc_cli:main", "fips-certs=fips_cli:main"]},
)



from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from sec_certs.dataset.cpe import CPEDataset
from sec_certs.sample.cpe import CPE


class TestCPE:
    @pytest.mark.slow
    @pytest.mark.monitor_test
    def test_from_web(self):
        with TemporaryDirectory() as tmpdir:
            dset = CPEDataset.from_web(Path(tmpdir) / "cpe.json")
        assert dset is not None
        assert "cpe:2.3:o:infineon:trusted_platform_firmware:6.40:*:*:*:*:*:*:*" in dset.cpes

    def test_from_to_dict(self):
        data = {
            "uri": "cpe:2.3:o:freebsd:freebsd:1.0:*:*:*:*:*:*:*",
            "title": None,
            "start_version": None,
            "end_version": None,
        }
        cpe = CPE.from_dict(data)
        ret = cpe.to_dict()
        assert data == ret
        other_cpe = CPE.from_dict(ret)
        assert cpe == other_cpe



import time
from multiprocessing.pool import ThreadPool
from typing import Any, Callable, Iterable, List, Optional, Union

from billiard.pool import Pool

from sec_certs.helpers import tqdm


def process_parallel(
    func: Callable,
    items: Iterable,
    max_workers: int,
    callback: Optional[Callable] = None,
    use_threading: bool = True,
    progress_bar: bool = True,
    unpack: bool = False,
    progress_bar_desc: Optional[str] = None,
) -> List[Any]:

    pool: Union[Pool, ThreadPool] = ThreadPool(max_workers) if use_threading else Pool(max_workers)
    results = (
        [pool.apply_async(func, (*i,), callback=callback) for i in items]
        if unpack
        else [pool.apply_async(func, (i,), callback=callback) for i in items]
    )

    if progress_bar is True and items:
        bar = tqdm(total=len(results), desc=progress_bar_desc)
        while not all(all_done := [x.ready() for x in results]):
            done_count = len(list(filter(lambda x: x, all_done)))
            bar.update(done_count - bar.n)
            time.sleep(1)
        bar.update(len(results) - bar.n)
        bar.close()

    pool.close()
    pool.join()
    pool.terminate()

    return [r.get() for r in results]
